from typing import Type, Union, List, Tuple, Dict

import pandas as pd
import numpy as np
from pandas import testing as tm
import sarracen
import pytest
import tempfile

from sarracen import SarracenDataFrame


def _create_capture_pattern(def_int: Type[np.generic],
                            def_real: Type[np.generic],
                            swap_endian: bool = False) -> bytearray:
    """ Construct capture pattern. """

    read_tag = np.array([13], dtype='int32')
    i1: np.ndarray = np.array([60769], dtype=def_int)
    r2: np.ndarray = np.array([60878], dtype=def_real)
    i2: np.ndarray = np.array([60878], dtype=def_int)
    iversion: np.ndarray = np.array([0], dtype=def_int)
    i3: np.ndarray = np.array([690706], dtype=def_int)

    if swap_endian:
        read_tag = read_tag.byteswap()
        i1 = i1.byteswap()
        r2 = r2.byteswap()
        i2 = i2.byteswap()
        iversion = iversion.byteswap()
        i3 = i3.byteswap()

    capture_pattern = bytearray(read_tag.tobytes())
    capture_pattern += bytearray(i1.tobytes())
    capture_pattern += bytearray(r2.tobytes())
    capture_pattern += bytearray(i2.tobytes())
    capture_pattern += bytearray(iversion.tobytes())
    capture_pattern += bytearray(i3.tobytes())
    capture_pattern += bytearray(read_tag.tobytes())

    return capture_pattern


def _create_file_identifier(swap_endian: bool = False) -> bytearray:
    """ Construct 100-character file identifier. """

    read_tag = np.array([13], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
    file_identifier = "Test of read_phantom".ljust(100)
    file = bytearray(read_tag.tobytes())
    file += bytearray(map(ord, file_identifier))
    file += bytearray(read_tag.tobytes())
    return file


def _create_global_header(massoftype: float = 1e-6,
                          massoftype_7: Union[float, None] = None,
                          def_int: Type[np.generic] = np.int32,
                          def_real: Type[np.generic] = np.float64,
                          mpi_blocks: int = 1,
                          swap_endian: bool = False) -> bytearray:
    """ Construct global variables. Only massoftype in this example. """

    dtypes = [def_int, np.int8, np.int16, np.int32, np.int64,
              def_real, np.float32, np.float64]
    param_dicts: List[Dict] = [dict() for _ in dtypes]

    params_def_int = param_dicts[0]
    params_def_real = param_dicts[5]

    params_def_real['massoftype'] = np.array([massoftype], dtype=def_real)
    if massoftype_7 is not None:
        params_def_real['massoftype_7'] = np.array([massoftype_7],
                                                   dtype=def_real)

    params_def_int['nblocks'] = np.array([mpi_blocks], dtype=def_int)

    dtype_param_pairs: List[Tuple[Type, Dict]] = list(zip(dtypes, param_dicts))

    read_tag = np.array([13], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
    file = bytearray()
    for dtype, params in dtype_param_pairs:
        nvars = np.array([len(params)], dtype='int32')
        if swap_endian:
            nvars = nvars.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(nvars.tobytes())
        file += bytearray(read_tag.tobytes())

        if len(params) > 0:
            file += bytearray(read_tag.tobytes())
            for k in params.keys():
                file += bytearray(map(ord, k.ljust(16)))
            file += bytearray(read_tag.tobytes())

            file += bytearray(read_tag.tobytes())
            for v in params.values():
                v_np = np.array([v], dtype=dtype)
                if swap_endian:
                    v_np = v_np.byteswap()
                file += bytearray(v_np)
            file += bytearray(read_tag.tobytes())

    return file


def _create_particle_array(tag: str,
                           data: list,
                           dtype: Type[np.generic] = np.float64,
                           swap_endian: bool = False) -> bytearray:

    read_tag = np.array([13], dtype='int32')
    data_np = np.array(data, dtype=dtype)
    if swap_endian:
        read_tag = read_tag.byteswap()
        data_np = data_np.byteswap()

    file = bytearray(read_tag.tobytes())
    file += bytearray(map(ord, tag.ljust(16)))
    file += bytearray(read_tag.tobytes())
    file += bytearray(read_tag.tobytes())
    file += bytearray(data_np.tobytes())
    file += bytearray(read_tag.tobytes())
    return file


@pytest.mark.parametrize("def_int, def_real, swap_endian",
                         [(np.int32, np.float64, False),
                          (np.int32, np.float32, False),
                          (np.int64, np.float64, False),
                          (np.int64, np.float32, False),
                          (np.int32, np.float64, True),
                          (np.int32, np.float32, True),
                          (np.int64, np.float64, True),
                          (np.int64, np.float32, True)])
def test_determine_default_precision(def_int: Type[np.generic],
                                     def_real: Type[np.generic],
                                     swap_endian: bool) -> None:
    """ Test if default int / real precision can be determined. """

    file = _create_capture_pattern(def_int, def_real, swap_endian)
    file += _create_file_identifier(swap_endian)
    file += _create_global_header(def_int=def_int, def_real=def_real,
                                  swap_endian=swap_endian)

    # create 1 block for gas
    read_tag = np.array([13], dtype='int32')
    nblocks = np.array([1], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
        nblocks = nblocks.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(nblocks.tobytes())
    file += bytearray(read_tag.tobytes())

    # 2 particles storing 1 default int and real arrays
    n = np.array([2], dtype='int64')
    nums = np.array([1, 0, 0, 0, 0, 1, 0, 0], dtype='int32')
    if swap_endian:
        n = n.byteswap()
        nums = nums.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(n.tobytes())
    file += bytearray(nums.tobytes())
    file += bytearray(read_tag.tobytes())

    # write particle arrays
    file += _create_particle_array("def_int", [1, 2],
                                   def_int, swap_endian)
    file += _create_particle_array("def_real", [1.0, 2.0],
                                   def_real, swap_endian)

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(file)
        fp.seek(0)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)

        assert list(sdf.dtypes) == [def_int, def_real]


@pytest.mark.parametrize("mpi_blocks, swap_endian",
                         [(1, False), (1, True),
                          (2, False), (2, True),
                          (4, False), (4, True)])
def test_gas_particles_only(mpi_blocks: int, swap_endian: bool) -> None:

    def_int = np.int32
    def_real = np.float64
    file = _create_capture_pattern(def_int, def_real, swap_endian)
    file += _create_file_identifier(swap_endian)
    file += _create_global_header(mpi_blocks=mpi_blocks,
                                  swap_endian=swap_endian)

    # create block for gas (broken into number of mpi_blocks)
    read_tag = np.array([13], dtype='int32')
    nblocks = np.array([mpi_blocks], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
        nblocks = nblocks.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(nblocks.tobytes())
    file += bytearray(read_tag.tobytes())

    x = [0, 0, 0, 0, 1, 1, 1, 1]
    y = [0, 0, 1, 1, 0, 0, 1, 1]
    z = [0, 1, 0, 1, 0, 1, 0, 1]
    h = [1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1]

    for i in range(mpi_blocks):
        # block header
        # 8 particles storing 4 real arrays (x, y, z, h)
        n = np.array([8 / mpi_blocks], dtype='int64')
        nums = np.array([0, 0, 0, 0, 0, 4, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        # block particle arrays
        # each mpi_block writes a chunk of the array
        size = len(x) // mpi_blocks
        start = i * size
        end = (i + 1) * size
        file += _create_particle_array("x", x[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("y", y[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("z", z[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("h", h[start:end],
                                       def_real, swap_endian)

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(file)
        fp.seek(0)

        sdf = sarracen.read_phantom(fp.name, separate_types='all')
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types='sinks')
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x),
                               check_index=False, check_names=False,
                               check_dtype=False)


@pytest.mark.parametrize("mpi_blocks, swap_endian",
                         [(1, False), (1, True),
                          (2, False), (2, True),
                          (4, False), (4, True)])
def test_gas_dust_particles(mpi_blocks: int, swap_endian: bool) -> None:

    def_int = np.int32
    def_real = np.float64
    file = _create_capture_pattern(def_int, def_real, swap_endian)
    file += _create_file_identifier(swap_endian)
    file += _create_global_header(massoftype_7=1e-4, mpi_blocks=mpi_blocks,
                                  swap_endian=swap_endian)

    # create block for gas & dust particles
    read_tag = np.array([13], dtype='int32')
    nblocks = np.array([mpi_blocks], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
        nblocks = nblocks.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(nblocks.tobytes())
    file += bytearray(read_tag.tobytes())

    itype = [1, 1, 1, 1, 1, 1, 1, 1,
             7, 7, 7, 7, 7, 7, 7, 7]
    x = [0, 0, 0, 0, 1, 1, 1, 1,
         0.5, 0.5, 0.5, 0.5, 1.5, 1.5, 1.5, 1.5]
    y = [0, 0, 1, 1, 0, 0, 1, 1,
         0.5, 0.5, 1.5, 1.5, 0.5, 0.5, 1.5, 1.5]
    z = [0, 1, 0, 1, 0, 1, 0, 1,
         0.5, 1.5, 0.5, 1.5, 0.5, 1.5, 0.5, 1.5]
    h = [1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1,
         1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1]

    for i in range(mpi_blocks):
        # block header
        # 8 particles storing 4 real arrays (x, y, z, h)
        n = np.array([16 / mpi_blocks], dtype='int64')
        nums = np.array([0, 1, 0, 0, 0, 4, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        # block particle arrays
        # each mpi_block writes a chunk of the array
        size = len(x) // mpi_blocks
        start = i * size
        end = (i + 1) * size
        file += _create_particle_array("itype", itype[start:end],
                                       np.int8, swap_endian)
        file += _create_particle_array("x", x[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("y", y[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("z", z[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("h", h[start:end],
                                       def_real, swap_endian)

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(file)
        fp.seek(0)

        sdf_g, sdf_d = sarracen.read_phantom(fp.name, separate_types='all')
        assert isinstance(sdf_g, SarracenDataFrame)
        assert isinstance(sdf_d, SarracenDataFrame)
        assert sdf_g.params is not None
        assert sdf_d.params is not None
        assert sdf_g.params['massoftype'] == 1e-6
        assert sdf_g.params['massoftype_7'] == 1e-4
        assert sdf_g.params['mass'] == 1e-6
        assert sdf_d.params['massoftype'] == 1e-6
        assert sdf_d.params['massoftype_7'] == 1e-4
        assert sdf_d.params['mass'] == 1e-4
        assert 'mass' not in sdf_g.columns
        assert 'mass' not in sdf_d.columns
        tm.assert_series_equal(sdf_g['x'],
                               pd.Series([0, 0, 0, 0, 1, 1, 1, 1]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_d['x'],
                               pd.Series([0.5, 0.5, 0.5, 0.5,
                                          1.5, 1.5, 1.5, 1.5]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types='sinks')
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf.params
        assert 'mass' in sdf.columns
        assert sdf[sdf.itype == 1]['mass'].unique() == [1e-6]
        assert sdf[sdf.itype == 7]['mass'].unique() == [1e-4]
        tm.assert_series_equal(sdf[sdf.itype == 1]['x'],
                               pd.Series([0, 0, 0, 0, 1, 1, 1, 1]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf[sdf.itype == 7]['x'],
                               pd.Series([0.5, 0.5, 0.5, 0.5,
                                          1.5, 1.5, 1.5, 1.5]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf.params
        assert 'mass' in sdf.columns
        assert sdf[sdf.itype == 1]['mass'].unique() == [1e-6]
        assert sdf[sdf.itype == 7]['mass'].unique() == [1e-4]
        tm.assert_series_equal(sdf[sdf.itype == 1]['x'],
                               pd.Series([0, 0, 0, 0, 1, 1, 1, 1]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf[sdf.itype == 7]['x'],
                               pd.Series([0.5, 0.5, 0.5, 0.5, 1.5,
                                          1.5, 1.5, 1.5]),
                               check_index=False, check_names=False,
                               check_dtype=False)


@pytest.mark.parametrize("mpi_blocks, swap_endian",
                         [(1, False), (1, True),
                          (2, False), (2, True),
                          (4, False), (4, True)])
def test_gas_sink_particles(mpi_blocks: int, swap_endian: bool) -> None:

    def_int = np.int32
    def_real = np.float64
    file = _create_capture_pattern(def_int, def_real, swap_endian)
    file += _create_file_identifier(swap_endian)
    file += _create_global_header(mpi_blocks=mpi_blocks,
                                  swap_endian=swap_endian)

    # block 1 = gas, block 2 = sinks
    read_tag = np.array([13], dtype='int32')
    nblocks = np.array([2 * mpi_blocks], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
        nblocks = nblocks.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(nblocks.tobytes())
    file += bytearray(read_tag.tobytes())

    x = [0, 0, 0, 0, 1, 1, 1, 1]
    y = [0, 0, 1, 1, 0, 0, 1, 1]
    z = [0, 1, 0, 1, 0, 1, 0, 1]
    h = [1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1]

    for i in range(mpi_blocks):
        # block headers
        n = np.array([8 / mpi_blocks], dtype='int64')
        nums = np.array([0, 0, 0, 0, 0, 4, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        n = np.array([1], dtype='int64')
        nums = np.array([0, 0, 0, 0, 0, 7, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        # write 4 gas particle arrays in block 1
        # each mpi_block writes a chunk of the array
        size = len(x) // mpi_blocks
        start = i * size
        end = (i + 1) * size
        file += _create_particle_array("x", x[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("y", y[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("z", z[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("h", h[start:end],
                                       def_real, swap_endian)

        # write 7 sink particle arrays in block 2
        # each mpi_block writes all sink particles (I believe)
        file += _create_particle_array("x", [0.000305],
                                       def_real, swap_endian)
        file += _create_particle_array("y", [-0.035809],
                                       def_real, swap_endian)
        file += _create_particle_array("z", [-0.000035],
                                       def_real, swap_endian)
        file += _create_particle_array("h", [1.0],
                                       def_real, swap_endian)
        file += _create_particle_array("spinx", [-3.911744e-8],
                                       def_real, swap_endian)
        file += _create_particle_array("spiny", [-1.326062e-8],
                                       def_real, swap_endian)
        file += _create_particle_array("spinz", [0.00058],
                                       def_real, swap_endian)

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(file)
        fp.seek(0)

        sdf, sdf_sinks = sarracen.read_phantom(fp.name, separate_types='all')
        assert isinstance(sdf, SarracenDataFrame)
        assert isinstance(sdf_sinks, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf_sinks.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf_sinks.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf_sinks.params
        assert 'mass' not in sdf.columns
        assert 'mass' not in sdf_sinks.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['x'], pd.Series([0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinx'], pd.Series([-3.911744e-8]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf, sdf_sinks = sarracen.read_phantom(fp.name, separate_types='sinks')
        assert isinstance(sdf, SarracenDataFrame)
        assert isinstance(sdf_sinks, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf_sinks.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf_sinks.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf_sinks.params
        assert 'mass' not in sdf.columns
        assert 'mass' not in sdf_sinks.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['x'], pd.Series([0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinx'], pd.Series([-3.911744e-8]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)
        assert isinstance(sdf, SarracenDataFrame)
        assert isinstance(sdf_sinks, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf_sinks.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['mass'] == 1e-6
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], pd.Series(x + [0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['h'], pd.Series(h + [1.0]),
                               check_index=False, check_names=False,
                               check_dtype=False)


@pytest.mark.parametrize("mpi_blocks, swap_endian",
                         [(1, False), (1, True),
                          (2, False), (2, True),
                          (4, False), (4, True)])
def test_gas_dust_sink_particles(mpi_blocks: int, swap_endian: bool) -> None:

    def_int = np.int32
    def_real = np.float64
    file = _create_capture_pattern(def_int, def_real, swap_endian)
    file += _create_file_identifier(swap_endian)
    file += _create_global_header(massoftype_7=1e-4, mpi_blocks=mpi_blocks,
                                  swap_endian=swap_endian)

    # create 1 block for gas
    read_tag = np.array([13], dtype='int32')
    nblocks = np.array([2 * mpi_blocks], dtype='int32')
    if swap_endian:
        read_tag = read_tag.byteswap()
        nblocks = nblocks.byteswap()
    file += bytearray(read_tag.tobytes())
    file += bytearray(nblocks.tobytes())
    file += bytearray(read_tag.tobytes())

    itype = [1, 1, 1, 1, 1, 1, 1, 1,
             7, 7, 7, 7, 7, 7, 7, 7]
    x = [0, 0, 0, 0, 1, 1, 1, 1,
         0.5, 0.5, 0.5, 0.5, 1.5, 1.5, 1.5, 1.5]
    y = [0, 0, 1, 1, 0, 0, 1, 1,
         0.5, 0.5, 1.5, 1.5, 0.5, 0.5, 1.5, 1.5]
    z = [0, 1, 0, 1, 0, 1, 0, 1,
         0.5, 1.5, 0.5, 1.5, 0.5, 1.5, 0.5, 1.5]
    h = [1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1,
         1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1]

    for i in range(mpi_blocks):
        # block headers
        # gas/dust particles in block 1
        n = np.array([16 / mpi_blocks], dtype='int64')
        nums = np.array([0, 1, 0, 0, 0, 4, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        # 1 sink particle in block 2
        n = np.array([1], dtype='int64')
        nums = np.array([0, 0, 0, 0, 0, 7, 0, 0], dtype='int32')
        if swap_endian:
            n = n.byteswap()
            nums = nums.byteswap()
        file += bytearray(read_tag.tobytes())
        file += bytearray(n.tobytes())
        file += bytearray(nums.tobytes())
        file += bytearray(read_tag.tobytes())

        # write gas/dust particle arrays in block 1
        # each mpi_block writes a chunk of the array
        size = len(x) // mpi_blocks
        start = i * size
        end = (i + 1) * size
        file += _create_particle_array("itype", itype[start:end],
                                       np.int8, swap_endian)
        file += _create_particle_array("x", x[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("y", y[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("z", z[start:end],
                                       def_real, swap_endian)
        file += _create_particle_array("h", h[start:end],
                                       def_real, swap_endian)

        # write 7 sink particle arrays in block 2
        # each mpi_block writes all sink particles (I believe)
        file += _create_particle_array("x", [0.000305],
                                       def_real, swap_endian)
        file += _create_particle_array("y", [-0.035809],
                                       def_real, swap_endian)
        file += _create_particle_array("z", [-0.000035],
                                       def_real, swap_endian)
        file += _create_particle_array("h", [1.0],
                                       def_real, swap_endian)
        file += _create_particle_array("spinx", [-3.911744e-8],
                                       def_real, swap_endian)
        file += _create_particle_array("spiny", [-1.326062e-8],
                                       def_real, swap_endian)
        file += _create_particle_array("spinz", [0.00058],
                                       def_real, swap_endian)

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(file)
        fp.seek(0)

        sdf_g, sdf_d, sdf_sinks = sarracen.read_phantom(fp.name,
                                                        separate_types='all')
        assert isinstance(sdf_g, SarracenDataFrame)
        assert isinstance(sdf_d, SarracenDataFrame)
        assert isinstance(sdf_sinks, SarracenDataFrame)
        assert sdf_g.params is not None
        assert sdf_d.params is not None
        assert sdf_sinks.params is not None
        assert sdf_g.params['massoftype'] == 1e-6
        assert sdf_g.params['massoftype_7'] == 1e-4
        assert sdf_g.params['mass'] == 1e-6
        assert sdf_d.params['massoftype'] == 1e-6
        assert sdf_d.params['massoftype_7'] == 1e-4
        assert sdf_d.params['mass'] == 1e-4
        assert sdf_sinks.params['massoftype'] == 1e-6
        assert sdf_sinks.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf_sinks.params
        assert 'mass' not in sdf_g.columns
        assert 'mass' not in sdf_d.columns
        assert 'mass' not in sdf_sinks.columns
        tm.assert_series_equal(sdf_g['x'],
                               pd.Series([0, 0, 0, 0, 1, 1, 1, 1]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_d['x'],
                               pd.Series([0.5, 0.5, 0.5, 0.5,
                                          1.5, 1.5, 1.5, 1.5]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['x'], pd.Series([0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinx'], pd.Series([-3.911744e-8]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf, sdf_sinks = sarracen.read_phantom(fp.name,
                                               separate_types='sinks')
        assert isinstance(sdf, SarracenDataFrame)
        assert isinstance(sdf_sinks, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf_sinks.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf.params
        assert 'mass' in sdf.columns
        assert sdf[sdf.itype == 1]['mass'].unique() == [1e-6]
        assert sdf[sdf.itype == 7]['mass'].unique() == [1e-4]
        tm.assert_series_equal(sdf[sdf.itype == 1]['x'],
                               pd.Series([0, 0, 0, 0, 1, 1, 1, 1]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf[sdf.itype == 7]['x'],
                               pd.Series([0.5, 0.5, 0.5, 0.5,
                                          1.5, 1.5, 1.5, 1.5]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        assert sdf_sinks.params['massoftype'] == 1e-6
        assert sdf_sinks.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf_sinks.params
        assert 'mass' not in sdf_sinks.columns
        tm.assert_series_equal(sdf_sinks['x'], pd.Series([0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinx'], pd.Series([-3.911744e-8]),
                               check_index=False, check_names=False,
                               check_dtype=False)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-6
        assert sdf.params['massoftype_7'] == 1e-4
        assert 'mass' not in sdf.params
        assert 'mass' in sdf.columns
        assert sdf[sdf.itype == 1]['mass'].unique() == [1e-6]
        assert sdf[sdf.itype == 7]['mass'].unique() == [1e-4]
        tm.assert_series_equal(sdf['x'], pd.Series([0, 0, 0, 0,
                                                    1, 1, 1, 1,
                                                    0.5, 0.5, 0.5, 0.5,
                                                    1.5, 1.5, 1.5, 1.5,
                                                    0.000305]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['h'], pd.Series([1.1] * 16 + [1.0]),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['mass'], pd.Series([1e-6] * 8
                                                      + [1e-4] * 8
                                                      + [np.nan]),
                               check_index=False, check_names=False,
                               check_dtype=False)



import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from sarracen import SarracenDataFrame
from sarracen.disc import surface_density
import pytest


def test_mass_equivalency() -> None:
    """ Column mass result should equal global params mass result. """

    # randomly place particles
    rng = np.random.default_rng(seed=5)
    x = rng.random(100)
    y = rng.random(100)
    z = rng.random(100)
    mass = [3.2e-4] * 100

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z, 'mass': mass})
    sigma1 = surface_density(sdf)

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z},
                            params={'mass': 3.2e-4})
    sigma2 = surface_density(sdf)

    assert_array_equal(sigma1, sigma2)


@pytest.mark.parametrize("geometry", ['cylindrical', 'spherical'])
def test_origin(geometry: str) -> None:
    """ Expect same profile regardless of origin. """

    rng = np.random.default_rng(seed=5)
    x = rng.random(100)
    y = rng.random(100)
    z = rng.random(100)

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z},
                            params={'mass': 3.2e-4})
    sigma_zero = surface_density(sdf, origin=[0.0, 0.0, 0.0],
                                 bins=30, geometry=geometry)

    sdf = SarracenDataFrame(data={'x': x - 2e-8,
                                  'y': y + 2e-8,
                                  'z': z + 3e-9},
                            params={'mass': 3.2e-4})
    sigma_eps = surface_density(sdf, origin=[2e-8, -2e-8, 3e-9],
                                bins=30, geometry=geometry)

    sdf = SarracenDataFrame(data={'x': x + 5.6e4,
                                  'y': y - 8.7e3,
                                  'z': z + 5.4e6},
                            params={'mass': 3.2e-4})
    sigma_large = surface_density(sdf, origin=[5.6e4, -8.7e3, 5.4e6],
                                  bins=30, geometry=geometry)

    assert_allclose(sigma_zero, sigma_eps, atol=2e-8, rtol=0.0)
    assert_allclose(sigma_zero, sigma_large, atol=2e-8, rtol=0.0)


def test_parts_vs_whole() -> None:
    """ Profiles should be the same for matching bins. """

    # randomly place particles
    rng = np.random.default_rng(seed=5)
    x = rng.random(100)
    y = rng.random(100)
    z = rng.random(100)
    vx = rng.random(100)
    vy = rng.random(100)
    vz = rng.random(100)
    mass = [3.2e-4] * 100

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z,
                                  'vx': vx, 'vy': vy, 'vz': vz,
                                  'mass': mass})
    sigma_in = surface_density(sdf, r_in=0.0, r_out=0.5, bins=100)
    sigma_out = surface_density(sdf, r_in=0.5, r_out=1.0, bins=100)
    sigma_all = surface_density(sdf, r_in=0.0, r_out=1.0, bins=200)

    assert_allclose(sigma_in, sigma_all[:100], atol=1e-15, rtol=0.0)
    assert_allclose(sigma_out, sigma_all[100:], atol=1e-15, rtol=0.0)



from .readers.read_csv import read_csv
from .readers.read_gasoline import read_gasoline
from .readers.read_gradsph import read_gradsph
from .readers.read_marisa import read_marisa
from .readers.read_phantom import read_phantom
from .readers.read_phantom_ev import read_phantom_ev
from .readers.read_shamrock import read_shamrock

try:
    from .readers.read_shamrock_vtk import read_shamrock_vtk
except ModuleNotFoundError as e:
    # temporary workaround for no VTK in Python 3.8 on Mac ARM
    if e.name == "vtk":
        def read_shamrock_vtk(*args, **kwargs):
            raise ImportError("vtk is not installed.")
    else:
        raise


from .writers.write_phantom import write_phantom

from .sarracen_dataframe import SarracenDataFrame

from .interpolate import interpolate_2d, interpolate_2d_line, \
    interpolate_3d_proj, interpolate_3d_cross
from .render import render, streamlines, arrowplot

from . import disc
from . import ptmass

__version__ = "1.3.1"

__all__ = ["read_csv", "read_gasoline", "read_gradsph", "read_marisa",
           "read_phantom", "read_phantom_ev", "read_shamrock",
           "read_shamrock_vtk", "write_phantom",
           "SarracenDataFrame", "disc", "ptmass",
           "interpolate_2d", "interpolate_2d_line", "interpolate_3d_proj",
           "interpolate_3d_cross", "render", "streamlines", "arrowplot"]



import numpy as np
import pandas as pd

import vtk
from vtk.util.numpy_support import vtk_to_numpy

from ..sarracen_dataframe import SarracenDataFrame


def read_shamrock_vtk(filename: str, pmass: float) -> SarracenDataFrame:
    """
    Read data from a Shamrock vtk file (compatible with Paraview).

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.
    pmass : float
        Mass of particles in the simulation (for now, it is assumed all
        particles have the same mass).

    Returns
    -------
    SarracenDataFrame

    """

    # read the vtk file
    reader = vtk.vtkDataSetReader()
    reader.SetFileName(filename)
    reader.Update()
    vtk_data = reader.GetOutput()
    num_arrays = vtk_data.GetPointData().GetNumberOfArrays()

    # initialize the dataframe
    df = pd.DataFrame()

    for i in range(num_arrays):
        vtk_array = vtk_data.GetPointData().GetArray(i)
        array_name = vtk_array.GetName()
        numpy_array = vtk_to_numpy(vtk_array)
        ndim = numpy_array.ndim

        if ndim == 1:
            df[array_name] = numpy_array
        else:
            df[array_name + 'x'] = numpy_array[:, 0]
            df[array_name + 'y'] = numpy_array[:, 1]
            df[array_name + 'z'] = numpy_array[:, 2]

    # add B
    if 'B/rhox' in df.columns:
        df['Bx'] = df['B/rhox'] * df['rho']
    if 'B/rhoy' in df.columns:
        df['By'] = df['B/rhoy'] * df['rho']
    if 'B/rhoz' in df.columns:
        df['Bz'] = df['B/rhoz'] * df['rho']

    # now add position columns
    points = vtk_data.GetPoints()
    numpy_points = vtk_to_numpy(points.GetData())

    df['x'] = numpy_points[:, 0]
    df['y'] = numpy_points[:, 1]
    df['z'] = numpy_points[:, 2]

    # finish by adding mass
    df['mass'] = pmass * np.ones_like(numpy_points[:, 0])
    return SarracenDataFrame(df)



import math

import numpy as np
from numba import njit, prange
from typing import Callable


class BaseKernel:
    """A generic kernel used for data interpolation."""

    def __init__(self) -> None:
        self._ckernel_func_cache = None
        self._column_cache = None

    @staticmethod
    def get_radius() -> float:
        """Get the smoothing radius of this kernel."""
        return 1

    @staticmethod
    def w(q: float, dim: int) -> float:
        """ Get the normalized weight of this kernel.

        Parameters
        ----------
        q : float
            The value to evaluate this kernel at.
        dim : {1, 2, 3}
            The number of dimensions to normalize the kernel value for.

        Returns
        -------
        float
            The normalized kernel weight at `q`.
        """

        return 1

    def get_column_kernel(self, samples: int = 1000) -> np.ndarray:
        """ Integrate a given 3D kernel over the z-axis.

        Parameters
        ----------
        samples: int
            Number of sample points to calculate when approximating the kernel.

        Returns
        -------
        A ndarray of length (samples), containing the kernel approximation.

        Examples
        --------
        Use np.linspace and np.interp to use this column kernel approximation:
            np.interp(q,
                      np.linspace(0, kernel.get_radius(), samples),
                      column_kernel)
        """
        if samples == 1000 and self._column_cache is not None:
            return self._column_cache

        c_kernel = BaseKernel._int_func(self.get_radius(), samples, self.w)

        if samples == 1000:
            self._column_cache = c_kernel

        return c_kernel

    def get_column_kernel_func(self, samples: int) -> Callable[[float, int],
                                                               float]:
        """ Generate a numba-accelerated column kernel function.

        Creates a numba-accelerated function for column kernel weights. This
        function can be utilized similarly to kernel.w().

        Parameters
        ----------
        samples: int
            Number of sample points to calculate when approximating the kernel.

        Returns
        -------
        A numba-accelerated weight function.
        """
        if self._ckernel_func_cache is not None and samples == 1000:
            return self._ckernel_func_cache
        column_kernel = self.get_column_kernel(samples)
        radius = self.get_radius()

        @njit(fastmath=True)
        def func(q: float, dim: int) -> float:
            # using np.linspace() would break compatibility with the GPU
            # backend, so the calculation here is performed manually.
            wab_index = q * (samples - 1) / radius
            index = min(max(0, int(math.floor(wab_index))), samples - 1)
            index1 = min(max(0, int(math.ceil(wab_index))), samples - 1)
            t = wab_index - index
            return column_kernel[index] * (1 - t) + column_kernel[index1] * t

        return func

    # Internal function for performing the integral in _get_column_kernel()
    @staticmethod
    @njit(fastmath=True, parallel=True)
    def _int_func(radius: float,
                  samples: int,
                  wfunc: Callable[[float, int], float]) -> np.ndarray:
        result = np.zeros(samples)

        for i in prange(samples):
            q_xy = radius * i / (samples - 1)
            bounds = np.sqrt(radius ** 2 - q_xy ** 2)
            q_z = np.linspace(0, bounds, samples)
            q = np.sqrt(q_xy ** 2 + q_z ** 2)
            y = wfunc(q, 3)
            result[i] = 2 * float(np.trapz(y, x=q_z))

        return result



from typing import Tuple, Union

import numpy as np
import pandas as pd
import os
import struct

from ..sarracen_dataframe import SarracenDataFrame


def read_gasoline(filename: str,
                  outtype: str = "sarracen") -> Union[Tuple[SarracenDataFrame,
                                                            SarracenDataFrame,
                                                            SarracenDataFrame],
                                                      Tuple[dict, dict,
                                                            dict, dict], int]:
    """
    Read data from a Gasoline tipsy file.

    Particles are separated into three data structures -- one for SPH gas
    particles, one for dark matter particles, and one for star particles.

    Global values stored in the dump file are stored within the
    SarracenDataFrames in the dictionary ``params``. If the output type is
    Python dictionaries, then an additional dictionary is returned with the
    global variables.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.
    outtype : {'sarracen', 'dict'}, default='sarracen'
        Whether to return data in SarracenDataFrames or Python dictionaries.

    Returns
    -------
    list of SarracenDataFrames or Python dictionaries

    Notes
    --------
    Adapted from `PyTipsy <https://github.com/bwkeller/pytipsy>`_.

    Examples
    --------
    Particles are separated into gas, dark matter, and star particles.

    >>> sdf_g, sdf_dm, sdf_star = sarracen.read_gasoline('dumpfile')

    The `outtype` parameter can be used to change the return type to Python
    dictionaries. This will return global variables in a separate dictionary.

    >>> header, catg, catd, cats = sarracen.read_gasoline('dumpfile', outtype='dict')
    """

    if outtype.lower() in ["dic", "dict", "dictionary"]:
        dictcheck = 1
    else:
        dictcheck = 0

    # Get all relevant files
    filin = filename.split("/")[-1]
    dirin = filename[0:len(filename)-len(filin)]
    if len(dirin) > 0:
        dirlist = os.listdir(dirin)
    else:
        dirlist = os.listdir()
        dirin = "."
    ii = 0
    while ii < len(dirlist):
        if filin not in dirlist[ii]:
            dirlist.remove(dirlist[ii])
        else:
            ii += 1
    ii = 0
    varlist = dirlist[1:]
    for var in dirlist:
        if var != filin:
            varlist[ii] = var[len(filin)+1:]
            ii += 1

    # From PyTipsy
    fp = open("{}/{}".format(dirin, filin), 'rb')
    fs = len(fp.read())
    fp.seek(0)

    # Take in the Header
    t, n, ndim, ng, nd, ns = struct.unpack("<diiiii", fp.read(28))
    endianswap = False

    # check Endianness
    if (ndim < 2 or ndim > 3):
        endianswap = True
        fp.seek(0)
        t, n, ndim, ng, nd, ns = struct.unpack(">diiiii", fp.read(28))

    # Catch for 4 byte padding
    if (fs == 32+48*ng+36*nd+44*ns):
        fp.read(4)
    # File is borked if this is true
    elif (fs != 28+48*ng+36*nd+44*ns):
        print("Tipsy ERROR: Header and file size inconsistend")
        print("Estimates: Header bytes: 28 or 32 (either is OK)")
        print("     ngas: ", ng, " bytes:", 48*ng)
        print("    ndark: ", nd, " bytes:", 38*nd)
        print("    nstar: ", ns, " bytes:", 44*ns)
        print("Actual File bytes:", fs, " does not work")
        fp.close()
        return 1

    # Make dicitonaries for data
    catg = {'mass': np.zeros(ng), 'pos': np.zeros((ng, 3)),
            'vel': np.zeros((ng, 3)), 'rho': np.zeros(ng),
            'tempg': np.zeros(ng), 'h_gas': np.zeros(ng),
            'zmetal': np.zeros(ng),  'phi': np.zeros(ng)}
    catd = {'mass': np.zeros(nd), 'pos': np.zeros((nd, 3)),
            'vel': np.zeros((nd, 3)), 'eps': np.zeros(nd),
            'phi': np.zeros(nd)}
    cats = {'mass': np.zeros(ns), 'pos': np.zeros((ns, 3)),
            'vel': np.zeros((ns, 3)), 'metals': np.zeros(ns),
            'tform': np.zeros(ns), 'eps': np.zeros(ns),
            'phi': np.zeros(ns)}
    for cat in ['g', 'd', 's']:
        j = 0
        for qty in ['x', 'y', 'z']:
            locals()['cat'+cat][qty] = locals()['cat'+cat]['pos'][:, j]
            locals()['cat'+cat]['v'+qty] = locals()['cat'+cat]['vel'][:, j]
            j += 1

    # Read in additional variables
    for var in varlist:
        fvar = open("{}.{}".format(filename, var), "rb")
        nfvar = len(fvar.read())
        fvar.seek(0)
        fvar.read(4)
        if int((nfvar-4)/4) == (ng+ns+nd):
            catg[var] = np.zeros(ng)
            catd[var] = np.zeros(nd)
            cats[var] = np.zeros(ns)
            if ng > 0:
                for i in range(ng):
                    if endianswap:
                        catg[var][i], = struct.unpack(">f", fvar.read(4))
                    else:
                        catg[var][i], = struct.unpack("<f", fvar.read(4))
            if ns > 0:
                for i in range(ns):
                    if endianswap:
                        cats[var][i], = struct.unpack(">f", fvar.read(4))
                    else:
                        cats[var][i], = struct.unpack("<f", fvar.read(4))
            if nd > 0:
                for i in range(nd):
                    if endianswap:
                        catd[var][i], = struct.unpack(">f", fvar.read(4))
                    else:
                        catd[var][i], = struct.unpack("<f", fvar.read(4))
        # read 3D variables
        elif int((nfvar-4)/12) == (ng+ns+nd):
            if dictcheck:
                catg[var] = np.zeros((ng, 3))
                catd[var] = np.zeros((nd, 3))
                cats[var] = np.zeros((ns, 3))
            varx = "{}x".format(var)
            vary = "{}y".format(var)
            varz = "{}z".format(var)
            catg[varx] = np.zeros(ng)
            catd[varx] = np.zeros(nd)
            cats[varx] = np.zeros(ns)
            catg[vary] = np.zeros(ng)
            catd[vary] = np.zeros(nd)
            cats[vary] = np.zeros(ns)
            catg[varz] = np.zeros(ng)
            catd[varz] = np.zeros(nd)
            cats[varz] = np.zeros(ns)
            if dictcheck:
                if (ng > 0):
                    for i in range(ng):
                        if endianswap:
                            catg[varx][i] = struct.unpack(">f", fvar.read(4))
                            catg[vary][i] = struct.unpack(">f", fvar.read(4))
                            catg[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            catg[varx][i] = struct.unpack("<f", fvar.read(4))
                            catg[vary][i] = struct.unpack("<f", fvar.read(4))
                            catg[varz][i] = struct.unpack("<f", fvar.read(4))
                        catg[var][i] = (catg[varx][i],
                                        catg[vary][i],
                                        catg[varz][i])
                if (ns > 0):
                    for i in range(ns):
                        if endianswap:
                            cats[varx][i] = struct.unpack(">f", fvar.read(4))
                            cats[vary][i] = struct.unpack(">f", fvar.read(4))
                            cats[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            cats[varx][i] = struct.unpack("<f", fvar.read(4))
                            cats[vary][i] = struct.unpack("<f", fvar.read(4))
                            cats[varz][i] = struct.unpack("<f", fvar.read(4))
                        catg[var][i] = (catd[varx][i],
                                        catd[vary][i],
                                        catd[varz][i])
                if (nd > 0):
                    for i in range(nd):
                        if endianswap:
                            catd[varx][i] = struct.unpack(">f", fvar.read(4))
                            catd[vary][i] = struct.unpack(">f", fvar.read(4))
                            catd[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            catd[varx][i] = struct.unpack("<f", fvar.read(4))
                            catd[vary][i] = struct.unpack("<f", fvar.read(4))
                            catd[varz][i] = struct.unpack("<f", fvar.read(4))
                        catg[var][i] = (cats[varx][i],
                                        cats[vary][i],
                                        cats[varz][i])
            else:
                if (ng > 0):
                    for i in range(ng):
                        if endianswap:
                            catg[varx][i] = struct.unpack(">f", fvar.read(4))
                            catg[vary][i] = struct.unpack(">f", fvar.read(4))
                            catg[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            catg[varx][i] = struct.unpack("<f", fvar.read(4))
                            catg[vary][i] = struct.unpack("<f", fvar.read(4))
                            catg[varz][i] = struct.unpack("<f", fvar.read(4))
                if (ns > 0):
                    for i in range(ns):
                        if endianswap:
                            cats[varx][i] = struct.unpack(">f", fvar.read(4))
                            cats[vary][i] = struct.unpack(">f", fvar.read(4))
                            cats[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            cats[varx][i] = struct.unpack("<f", fvar.read(4))
                            cats[vary][i] = struct.unpack("<f", fvar.read(4))
                            cats[varz][i] = struct.unpack("<f", fvar.read(4))
                if (nd > 0):
                    for i in range(nd):
                        if endianswap:
                            catd[varx][i] = struct.unpack(">f", fvar.read(4))
                            catd[vary][i] = struct.unpack(">f", fvar.read(4))
                            catd[varz][i] = struct.unpack(">f", fvar.read(4))
                        else:
                            catd[varx][i] = struct.unpack("<f", fvar.read(4))
                            catd[vary][i] = struct.unpack("<f", fvar.read(4))
                            catd[varz][i] = struct.unpack("<f", fvar.read(4))
        else:
            # Ignore Files outside of format
            print("Error With Variable {}".format(var))
        fvar.close()

    # Read in standard variables
    # Gas
    if (ng > 0):
        for i in range(ng):
            if endianswap:
                data = struct.unpack(">ffffffffffff", fp.read(48))
            else:
                data = struct.unpack("<ffffffffffff", fp.read(48))
            mass, x, y, z, vx, vy, vz, dens, tempg, h, zmetal, phi = data
            catg['mass'][i] = mass
            catg['x'][i] = x
            catg['y'][i] = y
            catg['z'][i] = z
            catg['vx'][i] = vx
            catg['vy'][i] = vy
            catg['vz'][i] = vz
            catg['rho'][i] = dens
            catg['tempg'][i] = tempg
            catg['h_gas'][i] = h
            catg['zmetal'][i] = zmetal
            catg['phi'][i] = phi

    # Dark Matter
    if (nd > 0):
        for i in range(nd):
            if endianswap:
                data = struct.unpack(">fffffffff", fp.read(36))
            else:
                data = struct.unpack("<fffffffff", fp.read(36))
            mass, x, y, z, vx, vy, vz, eps, phi = data
            catd['mass'][i] = mass
            catd['x'][i] = x
            catd['y'][i] = y
            catd['z'][i] = z
            catd['vx'][i] = vx
            catd['vy'][i] = vy
            catd['vz'][i] = vz
            catd['eps'][i] = eps
            catd['phi'][i] = phi

    # Stars
    if (ns > 0):
        for i in range(ns):
            if endianswap:
                data = struct.unpack(">fffffffffff", fp.read(44))
            else:
                data = struct.unpack("<fffffffffff", fp.read(44))
            mass, x, y, z, vx, vy, vz, metals, tform, eps, phi = data
            cats['mass'][i] = mass
            cats['x'][i] = x
            cats['y'][i] = y
            cats['z'][i] = z
            cats['vx'][i] = vx
            cats['vy'][i] = vy
            cats['vz'][i] = vz
            cats['metals'][i] = metals
            cats['tform'][i] = tform
            cats['eps'][i] = eps
            cats['phi'][i] = phi

    header = {'time': t, 'n': n, 'ndim': ndim,
              'ngas': ng, 'ndark': nd, 'nstar': ns}

    # Send dictionaries if desired
    if dictcheck:
        return header, catg, catd, cats
    # Otherwise, convert to PANDAS datframe
    else:
        catg.pop("pos")
        catg.pop("vel")
        catd.pop("pos")
        catd.pop("vel")
        cats.pop("pos")
        cats.pop("vel")
        if "smooth" not in varlist:
            catg['smooth'] = np.cbrt(catg['mass']/catg['rho'])
        catg['h'] = catg['smooth']
        dfg = pd.DataFrame.from_dict(catg)
        sdfg = SarracenDataFrame(dfg, params=header)
        dfd = pd.DataFrame.from_dict(catd)
        sdfd = SarracenDataFrame(dfd, params=header)
        dfs = pd.DataFrame.from_dict(cats)
        sdfs = SarracenDataFrame(dfs, params=header)
        return sdfg, sdfd, sdfs



from ..disc.surface_density import azimuthal_average, surface_density
from ..disc.surface_density import angular_momentum, scale_height, honH

__all__ = ["azimuthal_average", "surface_density", "angular_momentum",
           "scale_height", "honH"]



from typing import Any, Type, Union, Callable, Tuple, Optional, Dict, List

from matplotlib.axes import Axes
from matplotlib.colors import Colormap
import pandas as pd
from pandas import DataFrame, Series
from numba import cuda
import numpy as np
from scipy.spatial.transform import Rotation

from .render import streamlines, arrowplot, render, lineplot
from .interpolate import interpolate_2d, interpolate_3d_grid
from .kernels import CubicSplineKernel, BaseKernel


def _copy_doc(copy_func: Callable) -> Callable:
    """Copy documentation from another function to this function."""
    def wrapper(func: Callable) -> Callable:
        func.__doc__ = copy_func.__doc__
        return func
    return wrapper


class SarracenDataFrame(DataFrame):
    """
    A SarracenDataFrame is a pandas DataFrame with support for SPH data.

    A SarracenDataFrame is a subclass of the pandas DataFrame class designed to
    hold SPH particle data. Global simulation values are stored in ``params``,
    which is a standard Python dictionary.

    Interpolation and rendering functionality requires (at a minimum) particle
    positions, smoothing lengths and masses. SarracenDataFrames will attempt to
    identify columns which hold these data. For uniform, constant mass
    particles, the particle mass can be specified in the ``params`` dictionary.

    """

    _internal_names = pd.DataFrame._internal_names + ['_xcol', '_ycol',
                                                      '_zcol', '_mcol',
                                                      '_rhocol', '_hcol',
                                                      '_vxcol', '_vycol',
                                                      '_vzcol',
                                                      '_dustfracscol']
    _internal_names_set = set(_internal_names)

    _metadata = ['_params', '_units', '_kernel']

    def __init__(self, data=None, params=None, *args, **kwargs):
        """
        Construct a SarracenDataFrame from a NumPy array, dictionary, DataFrame
        or Iterable object.

        Parameters
        ----------
        data : ndarray, Iterable, DataFrame, or dict.
            Raw particle data which is passed to the pandas DataFrame
            constructor. Data can be specified in a dictionary, NumPy array or
            another DataFrame.
        params : dict, optional
            Global parameters from the simulation (time, hfact, etc). If
            constant, uniform mass particles are used, then the key ``mass``
            stores the particle mass (rather than specifying per particle).
        *args : tuple, optional
            Additional arguments to pass to the pandas DataFrame constructor.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the pandas DataFrame
            constructor.

        See Also
        --------
        :func:`read_csv` : Read data from a comma separated values (csv) file.
        :func:`read_phantom` : Read data from the Phantom SPH code.

        Examples
        --------
        Constructing using a Python dictionary.

        >>> particles = {'x': [1.0, 2.0, 3.0], 'y': [2.0, 2.0, 2.0], 'h': [3.0, 3.5, 4.0]}
        >>> sdf = sarracen.SarracenDataFrame(particles)
        >>> sdf
              x     y     h
        0   1.0   2.0   3.0
        1   2.0   2.0   3.5
        2   3.0   2.0   4.0

        Constructing using a two-dimensional NumPy array.

        >>> particles = np.array([[1.0, 2.0, 3.0], [2.0, 2.0, 3.5], [3.0, 2.0, 4.0]])
        >>> sdf = sarracen.SarracenDataFrame(particles, columns=['x', 'y', 'h'])
        >>> sdf
              x     y     h
        0   1.0   2.0   3.0
        1   2.0   2.0   3.5
        2   3.0   2.0   4.0

        Constant mass particles can specify mass in the ``params`` dictionary,
        rather than per particle.

        >>> particles = {'x': [1.0, 2.0, 3.0], 'y': [2.0, 2.0, 2.0], 'h': [3.0, 3.5, 4.0]}
        >>> params = {'mass': 0.2, 'hfact': 1.2}
        >>> sdf = sarracen.SarracenDataFrame(particles, params)
        >>> sdf.params
        {'mass': 0.2, 'hfact': 1.2}

        """

        # call pandas DataFrame constructor
        super().__init__(data, *args, **kwargs)

        self._params = dict(params or {})

        self._units = None
        self.units = Series([np.nan for _ in range(len(self.columns))])

        self._xcol, self._ycol, self._zcol = None, None, None
        self._hcol, self._mcol, self._rhocol = None, None, None
        self._vxcol, self._vycol, self._vzcol = None, None, None
        self._dustfracscol = []

        self._identify_special_columns()

        self._kernel = CubicSplineKernel()
        self._backend = 'gpu' if cuda.is_available() else 'cpu'

    @property
    def _constructor(self) -> Type:
        return SarracenDataFrame

    def _identify_special_columns(self) -> None:
        """
        Identify special columns commonly used in analysis functions.

        Identify which columns in this dataset correspond to important data
        columns commonly used in analysis functions. The columns which contain
        x, y, and z positional values are detected and set to the `xcol`,
        `ycol`, and `zcol` values. As well, the columns containing smoothing
        length, mass, and density information are identified and set to the
        `hcol`, `mcol`, and `rhocol`.

        If the x or y columns cannot be found, they are set to be the first two
        columns by default. If the z, smoothing length, mass, or density
        columns cannot be sound, the corresponding column label is set to
        `None`.
        """
        # First look for 'x', then 'rx', and then default to the first column.
        if 'x' in self.columns:
            self.xcol = 'x'
        elif 'rx' in self.columns:
            self.xcol = 'rx'
        elif len(self.columns) > 0:
            self.xcol = str(self.columns[0])

        # First look for 'y', then 'ry', and then default to the second column.
        if 'y' in self.columns:
            self.ycol = 'y'
        elif 'ry' in self.columns:
            self.ycol = 'ry'
        elif len(self.columns) > 1:
            self.ycol = str(self.columns[1])

        # First look for 'z', then 'rz', and then assume data is 2-dimensional.
        if 'z' in self.columns:
            self.zcol = 'z'
        elif 'rz' in self.columns:
            self.zcol = 'rz'

        # Look for the keyword 'h' in the data.
        if 'h' in self.columns:
            self.hcol = 'h'

        # Look for the keyword 'm' or 'mass' in the data.
        if 'm' in self.columns:
            self.mcol = 'm'
        elif 'mass' in self.columns:
            self.mcol = 'mass'

        # Look for the keyword 'rho' or 'density' in the data.
        if 'rho' in self.columns:
            self.rhocol = 'rho'
        elif 'density' in self.columns:
            self.rhocol = 'density'

        # Look for the keyword 'rho' or 'density' in the data.
        if 'vx' in self.columns:
            self.vxcol = 'vx'
        if 'vy' in self.columns:
            self.vycol = 'vy'
        if 'vz' in self.columns:
            self.vzcol = 'vz'

        # Look for the keyword 'dustfrac' in the data.
        for column in self.columns:
            if isinstance(column, str) and column.startswith('dustfrac'):
                self.dustfracscol.append(column)

    def calc_density(self) -> None:
        """
        Create a new column 'rho' that contains particle densities.

        Density for each particle is calculated according to

            .. math::

                \\rho = m \\left( \\frac{h_{\\rm fact}}{h}
                                \\right)^{n_{\\rm dim}}

        where :math:`m` is the particle mass, :math:`h` is the smoothing
        length, and :math:`h_{\\rm fact}` defines the ratio of smoothing length
        to particle spacing. Smoothing lengths are taken from the smoothing
        length column, particle masses from the mass column if present, or
        params if not, and hfact from params.

        Raises
        ------
        KeyError
            If the `hcol` column does not exist or if 'hfact' is not in
            'params'.
        ValueError
            If there is no particle mass data.

        Notes
        -----
        For one-fluid dump files, this will calculate the total density (gas +
        dust) of the particle.
        """
        if self.hcol not in self.columns:
            raise KeyError('Missing smoothing length data in this '
                           'SarracenDataFrame')

        if self.params is None or 'hfact' not in self.params:
            raise KeyError('hfact missing from params in this '
                           'SarracenDataFrame.')

        if self.mcol not in self.columns and (self.params is None or
                                              'mass' not in self.params):
            raise ValueError('Missing particle mass data in this '
                             'SarracenDataFrame.')

        # prioritize using mass per particle, if present
        if self.mcol in self.columns:
            mass = self[self.mcol]
        else:
            mass = self.params['mass']

        hfact = self.params['hfact']
        self['rho'] = mass * (hfact / self[self.hcol])**self.get_dim()
        self.rhocol = 'rho'

    def calc_one_fluid_quantities(self) -> None:
        """
        Create new columns that contain the densities of gas, dust (total),
        each dust grain size and the dust-to gas ratio in one-fluid
        (a.k.a. dust-as-mixture) simulations.

        Raises
        -------
        KeyError
            If `dustfrac` columns do not exist.
        ValueError
            If `ndustsmall` is zero or `ndustlarge` is non-zero.
        """
        if self.dustfracscol[0] not in self.columns:
            raise KeyError('Missing dust fraction data in this '
                           'SarracenDataFrame')

        if self.params['ndustsmall'] == 0 or self.params['ndustlarge'] != 0:
            raise ValueError('Not a one-fluid-only dump.')

        if self.rhocol not in self.columns:
            self.calc_density()

        if self.params['ndustsmall'] == 1:
            self['rho_g'] = self['rho'] * self['dustfrac']
            self['rho_d'] = self['rho'] * (1 - self['dustfrac'])
            self['dtg'] = self['rho_d'] / self['rho_g']
        else:
            self['dustfrac_total'] = self[self.dustfracscol].sum(axis=1)
            self['rho_g'] = self['rho'] * (1 - self['dustfrac_total'])
            self['rho_d_total'] = self['rho'] * self['dustfrac_total']
            self['rho_d'] = self['rho'] * self[self.dustfracscol[0]]
            for i in range(1, int(self.params['ndustsmall'])):
                self[f'rho_d_{i+1}'] = self['rho'] * self[self.dustfracscol[i]]
            self['dtg'] = self['dustfrac_total'] / (1 - self['dustfrac_total'])

    def centre_of_mass(self) -> list:
        """
        Returns the centre of mass of the data.

        Returns
        -------
        list
            A list with the centre of mass of the data.
        """
        if self.mcol in self.columns:
            mass = self[self.mcol]
        else:
            mass = self.params['mass']

        com_x = (self[self.xcol] * mass).sum()
        com_y = (self[self.ycol] * mass).sum()
        com_z = (self[self.zcol] * mass).sum()

        if isinstance(mass, pd.Series):
            inv_mass = 1.0 / mass.sum()
        else:
            inv_mass = 1.0 / (len(self) * mass)

        return [com_x * inv_mass, com_y * inv_mass, com_z * inv_mass]

    @_copy_doc(render)
    def render(self,
               target: str,
               x: Union[str, None] = None,
               y: Union[str, None] = None,
               z: Union[str, None] = None,
               xsec: Union[float, None] = None,
               kernel: Union[BaseKernel, None] = None,
               x_pixels: Union[int, None] = None,
               y_pixels: Union[int, None] = None,
               xlim: Union[Tuple[float, float], None] = None,
               ylim: Union[Tuple[float, float], None] = None,
               cmap: Union[str, Colormap] = 'gist_heat',
               cbar: bool = True,
               cbar_kws: dict = {},
               cbar_ax: Union[Axes, None] = None,
               ax: Union[Axes, None] = None,
               exact: bool = False,
               backend: Union[str, None] = None,
               integral_samples: int = 1000,
               rotation: Union[np.ndarray, list, Rotation, None] = None,
               rot_origin: Union[np.ndarray, list, str, None] = None,
               log_scale: bool = False,
               symlog_scale: bool = False,
               dens_weight: Union[bool, None] = None,
               normalize: bool = False,
               hmin: bool = False,
               **kwargs: Any) -> Axes:
        return render(self, target, x, y, z, xsec, kernel, x_pixels, y_pixels,
                      xlim, ylim, cmap, cbar, cbar_kws, cbar_ax, ax, exact,
                      backend, integral_samples, rotation, rot_origin,
                      log_scale, symlog_scale, dens_weight, normalize, hmin,
                      **kwargs)

    @_copy_doc(lineplot)
    def lineplot(self,
                 target: str,
                 x: Union[str, None] = None,
                 y: Union[str, None] = None,
                 z: Union[str, None] = None,
                 kernel: Union[BaseKernel, None] = None,
                 pixels: int = 512,
                 xlim: Union[Tuple[float, float], None] = None,
                 ylim: Union[Tuple[float, float], None] = None,
                 zlim: Union[Tuple[float, float], None] = None,
                 ax: Union[Axes, None] = None,
                 backend: Union[str, None] = None,
                 log_scale: bool = False,
                 dens_weight: bool = False,
                 normalize: bool = False,
                 hmin: bool = False,
                 **kwargs: Any) -> Axes:
        return lineplot(self, target, x, y, z, kernel, pixels, xlim, ylim,
                        zlim,  ax, backend, log_scale, dens_weight, normalize,
                        hmin, **kwargs)

    @_copy_doc(streamlines)
    def streamlines(self,
                    target: Union[Tuple[str, str], Tuple[str, str, str]],
                    x: Union[str, None] = None,
                    y: Union[str, None] = None,
                    z: Union[str, None] = None,
                    xsec: Union[int, None] = None,
                    kernel: Union[BaseKernel, None] = None,
                    integral_samples: int = 1000,
                    rotation: Union[np.ndarray, list, Rotation, None] = None,
                    rot_origin: Union[np.ndarray, list, str, None] = None,
                    x_pixels: Union[int, None] = None,
                    y_pixels: Union[int, None] = None,
                    xlim: Union[Tuple[float, float], None] = None,
                    ylim: Union[Tuple[float, float], None] = None,
                    ax: Union[Axes, None] = None,
                    exact: bool = False,
                    backend: Union[str, None] = None,
                    dens_weight: bool = False,
                    normalize: bool = False,
                    hmin: bool = False,
                    **kwargs: Any) -> Axes:
        return streamlines(self, target, x, y, z, xsec, kernel,
                           integral_samples, rotation, rot_origin, x_pixels,
                           y_pixels, xlim, ylim, ax, exact, backend,
                           dens_weight, normalize, hmin, **kwargs)

    @_copy_doc(arrowplot)
    def arrowplot(self,
                  target: Union[Tuple[str, str], Tuple[str, str, str]],
                  x: Union[str, None] = None,
                  y: Union[str, None] = None,
                  z: Union[str, None] = None,
                  xsec: Union[int, None] = None,
                  kernel: Union[BaseKernel, None] = None,
                  integral_samples: int = 1000,
                  rotation: Union[np.ndarray, list, Rotation, None] = None,
                  rot_origin: Union[np.ndarray, list, str, None] = None,
                  x_arrows: Union[int, None] = None,
                  y_arrows: Union[int, None] = None,
                  xlim: Union[Tuple[float, float], None] = None,
                  ylim: Union[Tuple[float, float], None] = None,
                  ax: Union[Axes, None] = None,
                  qkey: bool = True,
                  qkey_kws: Union[dict, None] = None,
                  exact: bool = False,
                  backend: Union[str, None] = None,
                  dens_weight: Union[bool, None] = None,
                  normalize: bool = False,
                  hmin: bool = False,
                  **kwargs: Any) -> Axes:
        return arrowplot(self, target, x, y, z, xsec, kernel, integral_samples,
                         rotation, rot_origin, x_arrows, y_arrows, xlim, ylim,
                         ax, qkey, qkey_kws, exact, backend, dens_weight,
                         normalize, hmin, **kwargs)

    def sph_interpolate(self,
                        target: str,
                        x: Union[str, None] = None,
                        y: Union[str, None] = None,
                        z: Union[str, None] = None,
                        kernel: Union[BaseKernel, None] = None,
                        rotation: Union[np.ndarray, list,
                                        Rotation, None] = None,
                        rot_origin: Union[np.ndarray, list, str, None] = None,
                        x_pixels: Union[int, None] = None,
                        y_pixels: Union[int, None] = None,
                        z_pixels: Union[int, None] = None,
                        xlim: Optional[Tuple[Optional[float],
                                             Optional[float]]] = None,
                        ylim: Optional[Tuple[Optional[float],
                                             Optional[float]]] = None,
                        zlim: Union[Tuple[float, float], None] = None,
                        exact: bool = False,
                        backend: str = 'cpu',
                        dens_weight: bool = False,
                        normalize: bool = False,
                        hmin: bool = False) -> np.ndarray:
        """
        Interpolate this data to a 2D or 3D grid, depending on the
        dimensionality of the data.

        Parameters
        ----------
        target: str
            The column label of the target data.
        x, y, z: str
            The column labels of the directional data to interpolate over.
            Defaults to the x, y, and z columns detected in `data`.
        kernel: BaseKernel
            The kernel to use for smoothing the target data. Defaults to the
            kernel specified in `data`.
        rotation: array_like or SciPy Rotation, optional
            The rotation to apply to the data before interpolation. If defined
            as an array, the order of rotations is [z, y, x] in degrees. Only
            applies to 3D datasets.
        rot_origin: array_like or ['com', 'midpoint'], optional
            Point of rotation of the data. Only applies to 3D datasets. If
            array_like, then the [x, y, z] coordinates specify the point around
            which the data is rotated. If 'com', then data is rotated around
            the centre of mass. If 'midpoint', then data is rotated around the
            midpoint, that is, min + max / 2. Defaults to the midpoint.
        x_pixels, y_pixels, z_pixels: int, optional
            Number of pixels in the output image in the x, y & z directions.
            Default values are chosen to keep a consistent aspect ratio.
        xlim, ylim, zlim: tuple of float, optional
            The minimum and maximum values to use in interpolation, in particle
            data space. Defaults to the minimum and maximum values of `x`, `y`
            and `z`.
        exact: bool
            Whether to use exact interpolation of the data. Only applies to
            2D datasets.
        backend: ['cpu', 'gpu']
            The computation backend to use when interpolating this data.
            Defaults to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A
            manually specified backend in `data` will override the default.
        dens_weight: bool
            If True, the target will be multiplied by density. Defaults to
            False.
        normalize: bool
            If True, will normalize the interpolation. Defaults to False (this
            may change in future versions).
        hmin: bool
            If True, a minimum smoothing length of 0.5 * pixel size will be
            imposed. This ensures each particle contributes to at least one
            grid cell / pixel. Defaults to False (this may change in a future
            verison).

        Returns
        -------
        ndarray (n-Dimensional)
            The interpolated output image, in a multi-dimensional numpy array.
            The number of dimensions match the dimensions of the data.
            Dimensions are structured in reverse order, where (x, y, z) ->
            [z, y, x].

        Raises
        -------
        ValueError
            If `x_pixels`, `y_pixels` or `z_pixels` are less than or equal to
            zero, or if the specified `x`, `y` and `z` minimum and maximum
            values result in an invalid region, or if `data` is not 2 or
            3-dimensional.
        KeyError
            If `target`, `x`, `y`, `z`, mass, density, or smoothing length
            columns do not exist in `data`.

        Examples
        --------

        Interpolate SPH particles to a uniform grid.

        >>> sdf, sdf_sinks = sarracen.read_phantom('dustydisc_00250')
        >>> grid = sdf.sph_interpolate('rho')

        The grid is a NumPy array in order of [z, y, x].

        The default dimensions of the grid are scaled to keep a similar aspect
        ratio. In this example, the z-direction is shorter because of the
        geometry of the data (accretion disc).

        >>> grid.shape
        (165,  526, 512)
        """
        if self.get_dim() == 2:
            if xlim is None:
                xlim = (None, None)
            if ylim is None:
                ylim = (None, None)
            return interpolate_2d(self, target, x, y, kernel, x_pixels,
                                  y_pixels, xlim, ylim, exact, backend,
                                  dens_weight, normalize, hmin)
        elif self.get_dim() == 3:
            return interpolate_3d_grid(self, target, x, y, z, kernel, rotation,
                                       rot_origin, x_pixels, y_pixels,
                                       z_pixels, xlim, ylim, zlim, backend,
                                       dens_weight, normalize, hmin)
        raise ValueError('Invalid number of dimensions.')

    @property
    def params(self) -> Dict[str, Any]:
        """
        dict: Miscellaneous dataset-level parameters.

        Raises
        ------
        TypeError
            If `params` is set to a non-dictionary or non-None object.
        """
        return self._params

    @params.setter
    def params(self, new_params: Union[Dict[str, Any], None]) -> None:
        if new_params is not None and not isinstance(new_params, dict):
            raise TypeError("Parameters not a dictionary")
        self._params = dict(new_params or {})

    @property
    def units(self) -> Series:
        """Series: Units for each column of this dataset."""
        return self._units

    @units.setter
    def units(self, new_units: Series) -> None:
        self._units = new_units

    @property
    def xcol(self) -> str:
        """
        str : Label of the column which contains x-positional data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._xcol

    @xcol.setter
    def xcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._xcol = new_col

    @property
    def ycol(self) -> str:
        """
        str : Label of the column which contains y-positional data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._ycol

    @ycol.setter
    def ycol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._ycol = new_col

    @property
    def zcol(self) -> str:
        """
        str : Label of the column which contains z-positional data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._zcol

    @zcol.setter
    def zcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._zcol = new_col

    @property
    def hcol(self) -> str:
        """
        str : Label of the column which contains smoothing length data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._hcol

    @hcol.setter
    def hcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._hcol = new_col

    @property
    def mcol(self) -> str:
        """
        str : Label of the column which contains particle mass data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._mcol

    @mcol.setter
    def mcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._mcol = new_col

    @property
    def rhocol(self) -> str:
        """
        str : Label of the column which contains particle density data.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._rhocol

    @rhocol.setter
    def rhocol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._rhocol = new_col

    @property
    def vxcol(self) -> str:
        """
        str : Label of the column which contains the x-component of the
        velocity.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._vxcol

    @vxcol.setter
    def vxcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._vxcol = new_col

    @property
    def vycol(self) -> str:
        """
        str : Label of the column which contains the y-component of the
        velocity.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._vycol

    @vycol.setter
    def vycol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._vycol = new_col

    @property
    def vzcol(self) -> str:
        """
        str : Label of the column which contains the z-component of the
        velocity.

        If this is set to a column which does not exist in the dataset, the
        column label will remain set to the old value.
        """
        return self._vzcol

    @vzcol.setter
    def vzcol(self, new_col: Union[str, None]) -> None:
        if new_col in self or new_col is None:
            self._vzcol = new_col

    @property
    def dustfracscol(self) -> List[str]:
        return self._dustfracscol

    @dustfracscol.setter
    def dustfracscol(self, new_col: List[str]) -> None:
        if new_col in self or new_col is None:
            self._dustfracscol = new_col

    @property
    def kernel(self) -> BaseKernel:
        """
        BaseKernel : The default kernel to use for interpolation operations
        with this dataset.

        If this is set to an object which is not a BaseKernel, the kernel will
        remain set as the old value.
        """
        return self._kernel

    @kernel.setter
    def kernel(self, new_kernel: BaseKernel) -> None:
        if isinstance(new_kernel, BaseKernel):
            self._kernel = new_kernel

    @property
    def backend(self) -> str:
        """
        ['cpu', 'gpu'] : The default backend to use for interpolation
        operations with this dataset.

        'cpu' - Best for small datasets, or cases where a GPU is not available.
        'gpu' - Best for large datasets, with a CUDA-enabled GPU.
        """
        return self._backend

    @backend.setter
    def backend(self, new_backend: str) -> None:
        self._backend = new_backend

    def get_dim(self) -> int:
        """
        Get the dimensionality of the data in this dataframe.

        Returns
        -------
        int
            The number of positional dimensions.
        """
        return 3 if self.zcol is not None else 2



import numpy as np
from numba import njit

from ..kernels import BaseKernel


class QuinticSplineKernel(BaseKernel):
    """An implementation of the Quintic Spline kernel."""

    @staticmethod
    def get_radius() -> float:
        return 3

    @staticmethod
    @njit(fastmath=True)
    def w(q: float, ndim: int) -> float:
        norm = 1 / 120 if (ndim == 1) \
            else 7 / (478 * np.pi) if (ndim == 2) \
            else 1 / (120 * np.pi)

        return norm * ((3 - q)**5 * (q < 3)
                       - 6 * (2 - q)**5 * (q < 2)
                       + 15 * (1 - q)**5 * (q < 1)) * (0 <= q)



"""pytest unit tests for render.py functions."""
from typing import Any, Callable, Dict, List
from matplotlib import pyplot as plt
from numba import cuda
from numpy.testing import assert_array_equal
from pytest import mark

import numpy as np

from sarracen import SarracenDataFrame
from sarracen import interpolate_2d, interpolate_2d_line
from sarracen import interpolate_3d_proj, interpolate_3d_cross
from sarracen.render import render, streamlines, arrowplot, lineplot

backends = ['cpu']
if cuda.is_available():
    backends.append('gpu')


@mark.parametrize("backend", backends)
def test_interpolation_passthrough(backend: str) -> None:
    """
    Verify that rendering functions use proper underlying interpolation.
    """
    data = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1],
            'Ax': [3, 2], 'Ay': [2, 1], 'h': [1, 1],
            'rho': [1, 1], 'm': [1, 1]}
    sdf = SarracenDataFrame(data)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', ax=ax)
    img = ax.images[0].get_array()
    assert img is not None
    assert_array_equal(img, interpolate_2d(sdf, 'P'))
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf, 'P', xlim=(3, 6), ylim=(1, 5), ax=ax)
    assert_array_equal(ax.lines[0].get_ydata(),
                       interpolate_2d_line(sdf, 'P', xlim=(3, 6), ylim=(1, 5)))
    plt.close(fig)

    data = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
            'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf = SarracenDataFrame(data)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', ax=ax)
    img = ax.images[0].get_array()
    assert img is not None
    assert_array_equal(img, interpolate_3d_proj(sdf, 'P'))
    plt.close(fig)

    fig, ax = plt.subplots()
    render(sdf, 'P', xsec=1.5, ax=ax)
    img = ax.images[0].get_array()
    assert img is not None
    assert_array_equal(img, interpolate_3d_cross(sdf, 'P'))
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_cmap(backend: str) -> None:
    """
    Verify that each rendering function uses the provided color map.
    """
    data = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1],
            'rho': [1, 1], 'm': [1, 1]}
    sdf = SarracenDataFrame(data)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)

    data = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
            'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf = SarracenDataFrame(data)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)

    fig, ax = plt.subplots()
    render(sdf, 'P', xsec=1.5, cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_cbar_exclusion(backend: str) -> None:
    """
    Verify that each rendering function respects the cbar argument.
    """
    data_2 = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_2 = SarracenDataFrame(data_2)
    sdf_2.backend = backend
    data_3 = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1],
              'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_3 = SarracenDataFrame(data_3)
    sdf_3.backend = backend

    args_list: List[Dict[str, Any]] = [{'data': sdf_2, 'xsec': None},
                                       {'data': sdf_3, 'xsec': None},
                                       {'data': sdf_3, 'xsec': 1.5}]
    for args in args_list:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], cbar=True, ax=ax)
        assert ax.images[-1].colorbar is not None
        plt.close(fig)

        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], cbar=False, ax=ax)
        assert ax.images[-1].colorbar is None
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_cbar_keywords(backend: str) -> None:
    """
    Verify that rendering functions respect the passed colorbar keywords.
    """
    data_2 = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_2 = SarracenDataFrame(data_2)
    sdf_2.backend = backend
    data_3 = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_3 = SarracenDataFrame(data_3)
    sdf_3.backend = backend

    args_list: List[Dict[str, Any]] = [{'data': sdf_2, 'xsec': None},
                                       {'data': sdf_3, 'xsec': None},
                                       {'data': sdf_3, 'xsec': 1.5}]
    for args in args_list:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'],
               cbar_kws={'orientation': 'horizontal'}, ax=ax)
        cb = ax.images[-1].colorbar
        assert cb is not None
        assert cb.orientation == 'horizontal'
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_kwargs(backend: str) -> None:
    """
    Verify that each rendering function respects passed keyword arguments.
    """
    data_2 = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
              'Ax': [1, 1], 'Ay': [1, 1]}
    sdf_2 = SarracenDataFrame(data_2)
    sdf_2.backend = backend
    data_3 = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_3 = SarracenDataFrame(data_3)
    sdf_3.backend = backend
    data_4 = {'x': [-3, 6], 'y': [5, -1], 'z': [2, 1], 'P': [-1, 1],
              'h': [1, 1], 'rho': [-1, -1], 'm': [1, 1]}
    sdf_4 = SarracenDataFrame(data_4)
    sdf_4.backend = backend

    args_list: List[Dict[str, Any]] = [{'data': sdf_2, 'xsec': None},
                                       {'data': sdf_3, 'xsec': None},
                                       {'data': sdf_3, 'xsec': 1.5}]
    for args in args_list:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax, origin='upper')
        assert ax.images[0].origin == 'upper'
        plt.close(fig)

    for arg in [True, False]:
        fig, ax = plt.subplots()
        render(sdf_4, 'P', ax=ax, log_scale=arg, symlog_scale=True,
               origin='upper', vmin=-1., vmax=1.)
        assert ax.images[0].origin == 'upper'
        plt.close(fig)

    fig, ax = plt.subplots()
    streamlines(sdf_2, ('Ax', 'Ay'), ax=ax, zorder=5)
    assert ax.patches[0].zorder == 5
    plt.close(fig)

    fig, ax = plt.subplots()
    arrowplot(sdf_2, ('Ax', 'Ay'), ax=ax, zorder=5)
    assert ax.collections[0].zorder == 5
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', xlim=(3, 6), ylim=(1, 5), ax=ax, linestyle='--')
    assert ax.lines[0].get_linestyle() == '--'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_plot_labels(backend: str) -> None:
    """
    Verify that plot labels for each rendering function are correct.
    """
    data_2 = {'x': [3, 6], 'y': [5, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
              'Ax': [1, 1], 'Ay': [1, 1]}
    sdf_2 = SarracenDataFrame(data_2)
    sdf_2.backend = backend

    data_3 = {'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
              'Ax': [1, 1], 'Ay': [1, 1], 'Az': [1, 1]}
    sdf_3 = SarracenDataFrame(data_3)
    sdf_3.backend = backend

    args_list: List[Dict[str, Any]] = [{'data': sdf_2, 'xsec': None},
                                       {'data': sdf_3, 'xsec': None},
                                       {'data': sdf_3, 'xsec': 0}]
    for args in args_list:
        column = args['data'] is sdf_3 and args['xsec'] is None

        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax)

        assert ax.get_xlabel() == 'x'
        assert ax.get_ylabel() == 'y'
        assert ax.figure.axes[1].get_ylabel() == \
               ('column ' if column else '') + 'P'
        plt.close(fig)

        fig, ax = plt.subplots()
        render(args['data'], 'rho', x='y', y='x', xsec=args['xsec'], ax=ax)

        assert ax.get_xlabel() == 'y'
        assert ax.get_ylabel() == 'x'
        assert ax.figure.axes[1].get_ylabel() == \
               ('column ' if column else '') + 'rho'
        plt.close(fig)

    functions_list: List[Callable] = [streamlines, arrowplot]
    for func in functions_list:
        fig, ax = plt.subplots()
        func(sdf_2, ('Ax', 'Ay'), ax=ax)

        assert ax.get_xlabel() == 'x'
        assert ax.get_ylabel() == 'y'
        plt.close(fig)

        fig, ax = plt.subplots()
        func(sdf_3, ('Ax', 'Ay', 'Az'), x='y', y='x', ax=ax)

        assert ax.get_xlabel() == 'y'
        assert ax.get_ylabel() == 'x'
        plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', ax=ax)

    assert ax.get_xlabel() == 'cross-section (x, y)'
    assert ax.get_ylabel() == 'P'
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'rho', x='y', y='x', ax=ax)

    assert ax.get_xlabel() == 'cross-section (y, x)'
    assert ax.get_ylabel() == 'rho'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_plot_bounds(backend: str) -> None:
    """
    Verify that plot bounds are set correctly for each rendering function.
    """
    data_2 = {'x': [6, 3], 'y': [5, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]}
    sdf_2 = SarracenDataFrame(data_2)
    sdf_2.backend = backend

    data_3 = {'x': [6, 3], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1],
              'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
              'Ax': [1, 1], 'Ay': [1, 1], 'Az': [1, 1]}
    sdf_3 = SarracenDataFrame(data_3)
    sdf_3.backend = backend

    args_list: List[Dict[str, Any]] = [{'data': sdf_2, 'xsec': None},
                                       {'data': sdf_3, 'xsec': None},
                                       {'data': sdf_3, 'xsec': 1.5}]
    for args in args_list:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax)

        assert ax.get_xlim() == (3, 6)
        assert ax.get_ylim() == (1, 5)

        if args['data'] is sdf_2:
            interpolate = interpolate_2d(sdf_2, 'P')
            assert ax.figure.axes[1].get_ylim() == (0, interpolate.max())
        else:
            if args['xsec']:
                interpolate = interpolate_3d_cross(sdf_3, 'P')
                assert ax.figure.axes[1].get_ylim() == (0, interpolate.max())
            else:
                interpolate = interpolate_3d_proj(sdf_3, 'P')
                assert ax.figure.axes[1].get_ylim() == (0, interpolate.max())
        plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', ax=ax)

    assert ax.get_xlim() == (0, 5)
    # 512 pixels across (by default), and both particles are in corners
    # therefore closest pixel to a particle is sqrt(41)/1024 units away
    # use default kernel to determine the max pressure value
    assert ax.get_ylim() == (0, interpolate_2d_line(sdf_2, 'P').max())
    plt.close(fig)

    functions_list: List[Callable] = [arrowplot, streamlines]
    for func in functions_list:
        fig, ax = plt.subplots()
        func(sdf_3, ('Ax', 'Ay', 'Az'), ax=ax)

        assert ax.get_xlim() == (3, 6)
        assert ax.get_ylim() == (1, 5)
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_log_vmin(backend: str) -> None:
    """
    Verify that the vmin value of log scales are set properly.
    """
    data = {'x': [6, 3, 2], 'y': [5, 1, 2], 'z': [3, 4, 2], 'P': [1, 1, 1],
            'h': [1, 1, 1], 'rho': [1e10, 1e7, 1e4], 'm': [1, 1, 1]}
    sdf = SarracenDataFrame(data)
    sdf.backend = backend

    # With no bounds defined, the log scale should cover 4 orders of magnitude.
    interpolate = interpolate_3d_proj(sdf, 'rho')
    fig, ax = plt.subplots()
    render(sdf, "rho", ax=ax, log_scale=True, x_pixels=20)

    imgs = ax.get_images()
    vmin, vmax = imgs[0].get_clim()

    assert vmin == 10 ** (np.log10(interpolate.max()) - 4)
    assert vmax == interpolate.max()

    # With only vmax defined, the vmin value is scaled accordingly.
    fig, ax = plt.subplots()
    render(sdf, "rho", vmax=10, log_scale=True, x_pixels=20)
    imgs = ax.get_images()
    vmin, vmax = imgs[0].get_clim()

    assert vmin == 1e-3
    assert vmax == 10

    # With both bounds defined, no adjustment is made.
    fig, ax = plt.subplots()
    render(sdf, "rho", vmin=5, vmax=10, log_scale=True, x_pixels=20)
    imgs = ax.get_images()
    vmin, vmax = imgs[0].get_clim()

    assert vmin == 5
    assert vmax == 10



from typing import IO, List, Dict, Any

import struct
import json
import numpy as np
import pandas as pd

from ..sarracen_dataframe import SarracenDataFrame


class FileReader:
    def __init__(self, file: IO) -> None:
        """
        Initialize the FileReader with a file object.

        Parameters
        ----------
        file : file object
            The file object to read from.

        Notes
        -----
        The file object is stored and the last read position is set to the
        current file position with `file.tell()`.
        """
        self.file = file
        self.last_position = file.tell()  # Keeps track of
        # the last read position

    def read_int64_and_string(self) -> str:
        # Read the 64-bit integer (8 bytes)
        """
        Reads a 64-bit integer and the following string from the file.
        The last position is updated to the current file position after
        the string read.

        Returns
        -------
        str
            The read string.
        """
        int_bytes = self.file.read(8)
        if len(int_bytes) != 8:
            raise ValueError("Failed to read 64-bit integer")

        # Unpack the 64-bit integer from the bytes
        length = struct.unpack("q", int_bytes)[0]  # 'q' is for
        # signed 64-bit integer

        # Read the string of 'length' bytes
        string_bytes = self.file.read(length)
        if len(string_bytes) != length:
            raise ValueError("Failed to read the full string")

        # Update the last position to the current
        # file pointer after the string read
        self.last_position = self.file.tell()

        # Decode the string from bytes to UTF-8 and return
        return string_bytes.decode("utf-8")

    def read_from_position(self, offset: int, num_bytes: int) -> bytes:
        """
        Read data from a specific position for a specified number of bytes.

        Parameters
        ----------
        offset : int
            The offset from the last read position to start reading.
        num_bytes : int
            The number of bytes to read from the file.

        Returns
        -------
        bytes
            The data read from the file.

        Raises
        ------
        ValueError
            If length of the data read does not correspond to
            the expected number of bytes.
        """
        # Calculate the absolute position to read from.
        position_to_seek = self.last_position + offset

        # Move to the specified position in the file
        self.file.seek(position_to_seek)

        # Now read the specified number of bytes
        data = self.file.read(num_bytes)
        if len(data) != num_bytes:
            raise ValueError("Failed to read the expected number of bytes")

        return data


def decode_bytes_to_doubles(byte_data: bytes) -> List[float]:
    """
    Decodes a byte array into a list of double precision floats.

    Parameters
    ----------
    byte_data : bytes
        The byte array to decode.

    Returns
    -------
    List[float]
        A list of double-precision floating-point numbers decoded from
        the byte array.

    Raises
    ------
    ValueError
        If the length of byte_data is not a multiple of 8.
    """
    # Ensure that the byte_data length is a multiple of 8 (since each
    # double is 8 bytes)
    if len(byte_data) % 8 != 0:
        raise ValueError("The length of the byte data must be a multiple of 8")

    num_doubles = len(byte_data) // 8  # Number of doubles in the byte array

    # Unpack the byte data into a tuple of doubles ('d' format
    # for double-precision floats)
    doubles = struct.unpack(f"{num_doubles}d", byte_data)

    # Return the list of double-precision floats
    return list(doubles)


def get_head_inc(off: int) -> int:
    if off % 8 > 0:
        off += 8 - (off % 8)
    return off


def decode_patchdata(pdat: bytes, pdat_layout: List[Dict[str, Any]]) -> dict:
    """
    Decode a patchdata bytearray into a dictionary of numpy arrays.

    Returns
    -------
    dic_out
        A dictionary with the decoded patchdata.
    """
    print(f"  Decoding patchdata with layout = {pdat_layout}")

    dic_out = {}

    pdl = pdat_layout

    lay_data = pdat[8:8 + len(pdat_layout) * 8]

    # the format start with the prehead lenght, so we skip it
    head = 0
    for i in range(len(pdat_layout)):
        field = pdat_layout[i]
        print(f"  Decoding {field} current head {head}")

        # because of direct GPU with use alignment 64
        nobj = struct.unpack("q", lay_data[head: head + 8])[0]
        print(f"    -> nobj = {nobj}")
        head += 8

        pdl[i]["nobj"] = nobj

    sz = 0
    for field in pdl:
        print(f"  Computing_size {field} current head {head}")
        nobj = field["nobj"]

        if field["type"] == "f64_3":
            sz += get_head_inc(8 * nobj * 3)
        elif field["type"] == "f64":
            sz += get_head_inc(8 * nobj)
        else:
            raise TypeError("No decoder for this type")

    head = 0
    pdat_dat = pdat[8 + len(pdat_layout) * 8:]

    print(sz, len(pdat_dat))

    for field in pdl:
        print(f"  Decoding {field} current head {head}")

        nobj = field["nobj"]

        if field["type"] == "f64_3":
            tmp = pdat_dat[head: head + 8 * nobj * 3]
            print(
                len(tmp),
                len(tmp) % 8,
                len(tmp) % 3,
                len(tmp) % (3 * 8),
                len(pdat_dat),
                head + 8 * nobj * 3,
            )
            data = decode_bytes_to_doubles(pdat_dat[head: head + 8 * nobj * 3])
            array_size = len(data)
            if array_size % (3 * nobj) != 0:
                raise ValueError(
                    f"Array size {array_size} is not equal to {3*nobj}"
                )
            else:
                dic_out[
                    field["field_name"]] = np.array(data).reshape((nobj, 3))

            head += get_head_inc(8 * nobj * 3)
        elif field["type"] == "f64":
            data = decode_bytes_to_doubles(pdat_dat[head: head + 8 * nobj])

            array_size = len(data)
            if array_size % nobj != 0:
                raise ValueError(
                    f"Array size {array_size} is not equal to {nobj}"
                    )

            dic_out[field["field_name"]] = np.array(data)

            head += get_head_inc(8 * nobj)
        else:
            raise TypeError("No decoder for this type")

    return dic_out


class ShamrockDumpReader:

    def __init__(self, file: IO) -> None:
        self.reader = FileReader(file)

        user_header = self.reader.read_int64_and_string()
        scheduler_header = self.reader.read_int64_and_string()
        filecontent_header = self.reader.read_int64_and_string()

        self.user_meta = json.loads(user_header)
        self.sched_meta = json.loads(scheduler_header)
        file_header = json.loads(filecontent_header)

        self.file_map = {}

        for bcount, off, pid in zip(
            file_header["bytecounts"],
            file_header["offsets"],
            file_header["pids"]
        ):
            self.file_map[pid] = {"bytecount": bcount, "offset": off}

    def read_patch(self, pid: np.uint64) -> dict:
        bcount = self.file_map[pid]["bytecount"]
        off = self.file_map[pid]["offset"]

        print(f"Reading patch pid={pid} (offset={off}, bytecount={bcount})")
        patchdata = self.reader.read_from_position(off, bcount)

        print(
            f"Decoding patchdata pid={pid} (offset={off}, "
            f"bytecount={bcount}), "
            f"len={len(patchdata)}"
        )

        return decode_patchdata(patchdata, self.sched_meta["patchdata_layout"])


def read_shamrock(filename: str) -> SarracenDataFrame:
    """
    Read data from a Shamrock native binary format dump file.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.

    Returns
    -------
    SarracenDataFrame

    Notes
    -----
    For now in the SPH solver there is just gas (no dust). Sinks information
    is not included in the dumps (see the Shamrock documentation
    https://shamrock-code.github.io/Shamrock/mkdocs/features/sph/sinks/)
    """
    full_df = []  # initialize the dataframe

    with open(filename, "rb") as f:
        reader = ShamrockDumpReader(f)

        # read metadata
        metadata = reader.user_meta
        mass = metadata["solver_config"]["gpart_mass"]
        metadata["mass"] = mass

        kernel = metadata["solver_config"]["kernel_id"]
        if kernel[:2] == "M4":
            hfact = 1.2
        elif kernel[:2] == "M5":
            hfact = 1.2
        elif kernel[:2] == "M6":
            hfact = 1.0
        else:
            raise KeyError("Unrecognised kernel.")
        metadata["hfact"] = hfact
        # end read metadata

        # read patches
        for pid in reader.file_map.keys():
            print(f"Reading patch {pid}")
            try:
                data = reader.read_patch(pid)
                patch_df = pd.DataFrame()

                for col in data.keys():
                    if data[col].ndim == 1:
                        if col == "hpart":
                            patch_df["h"] = data[col]

                        patch_df[col] = data[col]
                    else:
                        if col == "xyz":
                            patch_df["x"] = data[col][:, 0]
                            patch_df["y"] = data[col][:, 1]
                            patch_df["z"] = data[col][:, 2]
                        elif col == "vxyz":
                            patch_df["vx"] = data[col][:, 0]
                            patch_df["vy"] = data[col][:, 1]
                            patch_df["vz"] = data[col][:, 2]
                        else:
                            patch_df[col + "x"] = data[col][:, 0]
                            patch_df[col + "y"] = data[col][:, 1]
                            patch_df[col + "z"] = data[col][:, 2]

                full_df.append(patch_df)

            except Exception as e:
                print(f"Error reading patch {pid}: {e}")
                continue

    if not full_df:
        raise ValueError("No patch in this file!")

    final_df = pd.concat(
        full_df, ignore_index=True
    )  # combine in one df with coherent indexing

    return SarracenDataFrame(final_df, metadata)



import numpy as np
from numba import njit

from ..kernels import BaseKernel


class QuarticSplineKernel(BaseKernel):
    """An implementation of the Quartic Spline kernel."""

    @staticmethod
    def get_radius() -> float:
        return 2.5

    @staticmethod
    @njit(fastmath=True)
    def w(q: float, ndim: int) -> float:
        norm = 1 / 24 if (ndim == 1) \
            else 96 / (1199 * np.pi) if (ndim == 2) \
            else 1 / (20 * np.pi)

        return norm * ((2.5 - q)**4 * (q < 2.5)
                       - 5 * (1.5 - q)**4 * (q < 1.5)
                       + 10 * (0.5 - q)**4 * (q < 0.5)) * (0 <= q)



from typing import Dict, List, Union, Type, Tuple, Any
import numpy as np
import re
import warnings
from datetime import datetime

from ..sarracen_dataframe import SarracenDataFrame


params_dict_type = Dict[str, Union[np.generic, str]]


def _write_fortran_block(value: List[np.generic],
                         dtype: Type[np.generic]) -> bytearray:
    write_tag = np.array([len(value) * dtype().itemsize], dtype=np.int32)
    file = bytearray(write_tag.tobytes())
    file += bytearray(np.array(value, dtype=dtype).tobytes())
    file += bytearray(write_tag.tobytes())
    return file


def _write_file_identifier(sdf: SarracenDataFrame) -> bytearray:
    """ Write the file identifier.

    The file identifier is a 100-character string that encodes the various
    options that were used.

    First character is either 'F' or 'S', corresponding to Full or Small dump.
    Second character is 'T', for a tagged dump file.

    'Phantom' should appear in the name to declare it as a Phantom dump, not
    an sphNG dump file.
    """

    if 'file_identifier' not in sdf.params:
        now = datetime.now()
        frac = now.microsecond // 100000
        timestamp = f"{now:%d/%m/%Y %H:%M:%S}.{frac}"
        fileid = f"FT:Phantom:Sarracen: {timestamp}"
        sdf.params['file_identifier'] = fileid

    file_id = sdf.params['file_identifier'].ljust(100)
    file_id = list(map(ord, file_id))
    file_id = [np.uint8(c) for c in file_id]
    file = _write_fortran_block(file_id, dtype=np.uint8)
    return file


def _write_capture_pattern(def_int: Type[np.generic],
                           def_real: Type[np.generic],
                           iversion: int = 1) -> bytearray:
    write_tag = np.array([16 + def_real().itemsize], dtype=np.int32)
    i1 = np.array([60769], dtype=def_int)
    r2 = np.array([60878], dtype=def_real)
    i2 = np.array([60878], dtype=np.int32)
    iversion_arr = np.array([iversion], dtype=np.int32)
    i3 = np.array([690706], dtype=np.int32)

    capture_pattern = bytearray(write_tag.tobytes())
    capture_pattern += bytearray(i1.tobytes())
    capture_pattern += bytearray(r2.tobytes())
    capture_pattern += bytearray(i2.tobytes())
    capture_pattern += bytearray(iversion_arr.tobytes())
    capture_pattern += bytearray(i3.tobytes())
    capture_pattern += bytearray(write_tag.tobytes())

    return capture_pattern


def _check_for_essential_data(sdf: SarracenDataFrame) -> None:
    """ Check for missing essential columns or params."""

    required_cols = [sdf.xcol, sdf.ycol, sdf.zcol, sdf.hcol]
    if any(col is None for col in required_cols):
        missing: List[str] = []
        if sdf.xcol is None:
            missing.append("x")
        if sdf.ycol is None:
            missing.append("y")
        if sdf.zcol is None:
            missing.append("z")
        if sdf.hcol is None:
            missing.append("h")

        raise ValueError(f"Missing required columns: {', '.join(missing)}")

    units = ['udist', 'umass', 'utime', 'umagfd']
    if any(key not in sdf.params for key in units):
        msg = ("params is missing 'udist', 'umass', 'utime' and/or 'umagfd'. "
               "Phantom will assume default units.")
        warnings.warn(msg, UserWarning, stacklevel=3)

    ieos = sdf.params.get('ieos')

    if ieos == 1:
        if 'RK2' not in sdf.params:
            raise ValueError("'RK2' is required in params for 'ieos'==1")

    if ieos == 2:
        if 'gamma' not in sdf.params:
            raise ValueError("'gamma' is required in params for 'ieos'==2")

    if ieos == 3:
        # Phantom seems to run ok with this, but then assumes cs_min=0
        if 'cs_min' not in sdf.params:
            msg = ("'cs_min' not set in params even though 'ieos'==3.")
            warnings.warn(msg, UserWarning, stacklevel=3)
        # essential keys
        keys = ['RK2', 'qfacdisc']
        if any(key not in sdf.params for key in keys):
            msg = ("'RK2' and 'qfacdisc' are required in params for 'ieos'=3")
            raise ValueError(msg)


def _remove_invalid_keys(params: params_dict_type) -> params_dict_type:
    """ Remove keys specific to Sarracen internal use."""

    exclude = ['file_identifier', 'mass', 'def_int_dtype',
               'def_real_dtype', 'iversion']
    return {k: v for k, v in params.items() if k not in exclude}


def _standardize_dtypes(params: Dict[str, Any]) -> params_dict_type:
    """ Convert all params to numpy dtypes."""

    for k, v in params.items():
        if isinstance(v, int):
            params[k] = np.int32(v)
        if isinstance(v, float):
            params[k] = np.float64(v)

    return params


def _validate_ntypes(sdf: SarracenDataFrame,
                     params: params_dict_type) -> params_dict_type:
    """ Update params ntypes to reflect particle data."""

    if 'ntypes' in params:
        ntypes = int(params['ntypes'])

    elif 'itype' in sdf.columns:
        # guess using number of itypes
        ntypes = max(8, sdf['itype'].max())

    elif 'massoftype' in params:
        # guess using number of massoftype
        ntypes = 8
        for key in params:
            if key.startswith('massoftype_'):
                suffix = key[len('massoftype_'):]
                if suffix.isdigit():
                    ntypes = max(ntypes, int(suffix))

    elif 'npartoftype' in params:
        # guess using number of npartoftype
        # more difficult because npartoftype is stored as both int32 and int64
        # find max suffix for both, then divide the larger by 2
        max_int32 = 8
        max_int64 = 8
        for key in params:
            if key.startswith('npartoftype_'):
                suffix = key[len('npartoftype_'):]
                if suffix.isdigit():
                    if isinstance(params[key], np.int64):
                        max_int64 = max(int(suffix), max_int64)
                    else:
                        max_int32 = max(int(suffix), max_int32)
        # hack
        if max_int32 > max_int64:
            max_int32 = max_int32 // 2
        else:
            max_int64 = max_int64 // 2
        ntypes = max(8, max(max_int64, max_int32))

    else:
        ntypes = 8

    params['ntypes'] = np.int32(ntypes)
    params['ntypes_2'] = np.int64(ntypes)

    return params


def _validate_particle_counts(sdf: SarracenDataFrame,
                              params: params_dict_type) -> params_dict_type:
    """
    Update params particle counts to match actual particle counts.

    This will count the number of gas and dust particles based on itype. It
    assumes that if itype is not present, then all particles are gas.

    This will populate nparttot and npartoftype accordingly.
    """

    n_gas = len(sdf[sdf.itype == 1]) if 'itype' in sdf.columns else len(sdf)
    n_dust = len(sdf[sdf.itype == 7]) if 'itype' in sdf.columns else 0
    n_total = len(sdf)

    idust = params['idust'] if 'idust' in params else 7
    ntypes = int(params['ntypes'])

    # check total particle count
    if 'nparttot' not in params:
        params['nparttot'] = np.int32(n_total)
    if params['nparttot'] != n_total:
        params['nparttot'] = type(params['nparttot'])(n_total)

    # check int64 total particle count
    if 'nparttot_2' not in params:
        params['nparttot_2'] = np.int64(n_total)
    if params['nparttot_2'] != n_total:
        params['nparttot_2'] = type(params['nparttot_2'])(n_total)

    # set default values for npartoftypes if not set
    default = {'': n_gas}  # insertion order matters
    default.update({f'_{i}': 0 for i in range(2, ntypes + 1)})
    default[f'_{idust}'] = n_dust
    for i, value in default.items():
        key = f'npartoftype{i}'
        if key not in params:
            params[key] = np.int32(value)

    # set default values for int64 npartoftypes if not set
    default = {f'_{i}': 0 for i in range(ntypes + 1, 2 * ntypes + 1)}
    default[f'_{ntypes + 1}'] = n_gas
    default[f'_{ntypes + int(idust)}'] = n_dust
    for i, value in default.items():
        key = f'npartoftype{i}'
        if key not in params:
            params[key] = np.int64(value)

    # update gas and dust npartoftype if values are incorrect
    if params['npartoftype'] != n_gas:
        params['npartoftype'] = type(params['npartoftype'])(n_gas)
    key = f'npartoftype_{idust}'
    if params[key] != n_dust:
        params[key] = type(params[key])(n_dust)

    # update int64 gas and dust npartoftype if values are incorrect
    key = 'npartoftype_' + str(ntypes + 1)
    if params[key] != n_gas:
        params[key] = type(params[key])(n_gas)
    key = 'npartoftype_' + str(ntypes + int(idust))
    if params[key] != n_dust:
        params[key] = type(params[key])(n_dust)

    return params


def _validate_particle_masses(sdf: SarracenDataFrame,
                              params: params_dict_type) -> params_dict_type:
    """ Update params particle masses to match actual particle masses."""

    if 'mass' in params:
        m_gas = params['mass']
    elif 'massoftype' in params:
        m_gas = params['massoftype']
    elif sdf.mcol is not None and 'itype' not in sdf.columns:
        m_gas = sdf[sdf.mcol].iloc[0]
    elif sdf.mcol is not None and 'itype' in sdf.columns:
        m_gas = sdf[sdf.itype == 1][sdf.mcol].iloc[0]
    else:
        raise ValueError("Could not find gas mass.")

    # set default values for massoftype if not set
    default = {'': m_gas}  # insertion order matters
    tmp = {f'_{i}': np.float64(0) for i in range(2, int(params['ntypes']) + 1)}
    default.update(tmp)
    for i, value in default.items():
        key = f'massoftype{i}'
        if key not in params:
            params[key] = np.float64(value)

    idust = params['idust'] if 'idust' in params else 7

    # update massoftype for dust, if dust is present
    if 'itype' in sdf.columns and sdf.mcol is not None:
        m_dust_particle = sdf[sdf.itype == idust][sdf.mcol].iloc[0]

        if params[f'massoftype_{idust}'] == 0:
            params[f'massoftype_{idust}'] = m_dust_particle

        if params[f'massoftype_{idust}'] != m_dust_particle:
            raise ValueError("Mismatch in dust params and particle masses")

    return params


def _relocate_special_params(param_dicts: List[params_dict_type]) -> None:
    """Move certain parameters to their required dtype indices."""
    # Integer parameters that must be int32
    for param in ['iexternalforce', 'ieos']:
        if param in param_dicts[0]:
            param_dicts[3][param] = param_dicts[0].pop(param)

    # Real parameters that must be float64
    for param in ['udist', 'umass', 'utime', 'umagfd']:
        if param in param_dicts[5]:
            param_dicts[7][param] = param_dicts[5].pop(param)


def _rename_duplicate(tag: str) -> str:
    if len(tag) > 1 and tag[-2] == '_' and tag[-1].isdigit():
        tag = tag[:-2]
    if len(tag) > 2 and tag[-3] == '_' and tag[-2:].isdigit():
        tag = tag[:-3]

    return tag


def _write_global_header_array(tags: List[str],
                               values: List[np.generic],
                               dtype: Type[np.generic]) -> bytearray:
    tags = [_rename_duplicate(tag) for tag in tags]
    tags2: List[List[int]] = [list(map(ord, tag.ljust(16))) for tag in tags]
    tags3: List[np.uint8] = [np.uint8(c) for tag in tags2 for c in tag]

    file = _write_fortran_block(tags3, np.uint8)
    file += _write_fortran_block(values, dtype)

    return file


def sort_key(k: str) -> Tuple[str, int]:
    """
    Returns (basename, number) where basename is the part before any
    numeric suffix, and number is an integer used for sorting.
    Keys without numeric suffix get number = 0.
    """
    m = re.match(r"^(.*?)(?:_(\d+))?$", k)
    if m is None:
        raise ValueError(f"Invalid key format: {k!r}")

    base = m.group(1)
    num = int(m.group(2)) if m.group(2) else 0
    return (base, num)


def _reorder_params(params: params_dict_type) -> params_dict_type:
    return dict(sorted(params.items(), key=lambda item: sort_key(item[0])))


def _write_global_header(sdf: SarracenDataFrame,
                         def_int: Type[np.generic],
                         def_real: Type[np.generic]) -> bytearray:

    params_dict = sdf.params.copy()
    params_dict = _standardize_dtypes(params_dict)
    params_dict = _validate_ntypes(sdf, params_dict)
    params_dict = _validate_particle_counts(sdf, params_dict)
    params_dict = _validate_particle_masses(sdf, params_dict)
    params_dict = _reorder_params(params_dict)
    params_dict = _remove_invalid_keys(params_dict)

    dtypes = [def_int, np.int8, np.int16, np.int32, np.int64,
              def_real, np.float32, np.float64]

    # create params dict per dtype and populate
    param_dtype_dicts: List[params_dict_type] = [dict() for _ in dtypes]
    for k, v in params_dict.items():
        if isinstance(v, def_int):
            param_dtype_dicts[0][k] = v
        elif isinstance(v, np.int8):
            param_dtype_dicts[1][k] = v
        elif isinstance(v, np.int16):
            param_dtype_dicts[2][k] = v
        elif isinstance(v, np.int32):
            param_dtype_dicts[3][k] = v
        elif isinstance(v, np.int64):
            param_dtype_dicts[4][k] = v

        if isinstance(v, def_real):
            param_dtype_dicts[5][k] = v
        elif isinstance(v, np.float32):
            param_dtype_dicts[6][k] = v
        elif isinstance(v, np.float64):
            param_dtype_dicts[7][k] = v

    # some tags live in specific dtypes
    _relocate_special_params(param_dtype_dicts)

    # create header arrays per dtype
    header_data = []
    for i, dtype in enumerate(dtypes):
        tags = []
        values = []
        for k, v in param_dtype_dicts[i].items():
            if not isinstance(v, np.generic):
                raise ValueError(f"Invalid value type: {v!r}")
            tags.append(k)
            values.append(v)
        header_data.append((dtype, tags, values))

    file = bytearray()

    # write header arrays in dtype sequence
    for dtype, tags, values in header_data:
        nvars = np.int32(len(tags))
        file += _write_fortran_block([nvars], dtype=np.int32)

        if nvars > 0:
            file += _write_global_header_array(tags, values, dtype)

    return file


def _get_array_tags(sdf: SarracenDataFrame,
                    dtype: Type[np.generic]) -> List[str]:
    return list(sdf.select_dtypes(include=[dtype]).columns)


def _get_last_index(sdf: SarracenDataFrame) -> int:
    return 1 if sdf.index[-1] == 0 else sdf.shape[0]


def _write_array_blocks(sdf: SarracenDataFrame,
                        def_int: Type[np.generic],
                        def_real: Type[np.generic],
                        sinks: Union[SarracenDataFrame,
                                     None] = None) -> bytearray:

    dtypes = [def_int, np.int8, np.int16, np.int32, np.int64,
              def_real, np.float32, np.float64]

    nblocks = 2 if sinks is not None else 1
    file = _write_fortran_block([np.int32(nblocks)], np.int32)

    sdf_list = [sdf] if sinks is None else [sdf, sinks]
    sdf_dtype_info = []

    for sdf in sdf_list:
        nvars = np.array([_get_last_index(sdf)], dtype=np.int64)
        dtype_tags = []
        used = set()  # avoid def_int, def_real duplicates

        for dtype in dtypes:
            tags = _get_array_tags(sdf, dtype) if dtype not in used else []
            dtype_tags.append((dtype, tags))
            used.add(dtype)

        counts = np.array([len(tags) for _, tags in dtype_tags],
                          dtype=np.int32)
        write_tag = np.array([len(nvars) * nvars.dtype.itemsize
                              + len(counts) * counts.dtype.itemsize],
                             dtype=np.int32)

        # A mix of int64 and int32, so custom write
        file += write_tag.tobytes()
        file += nvars.tobytes()
        file += counts.tobytes()
        file += write_tag.tobytes()

        sdf_dtype_info.append((sdf, dtype_tags))

    for sdf, dtype_tags in sdf_dtype_info:
        for dtype, tags in dtype_tags:
            for tag in tags:
                # Phantom needs specifically 'x', 'y', etc tags
                # irregardless of what the xcol, ycol, etc are set to
                mapping = {sdf.xcol: 'x', sdf.ycol: 'y',
                           sdf.zcol: 'z', sdf.hcol: 'h'}
                write_tag = mapping.get(tag, tag)
                write_tag = _rename_duplicate(write_tag).ljust(16)
                file += _write_fortran_block(list(map(ord, write_tag)),
                                             dtype=np.uint8)
                file += _write_fortran_block(list(sdf[tag]), dtype)
    return file


def write_phantom(filename: str,
                  sdf: SarracenDataFrame,
                  sinks: Union[SarracenDataFrame, None] = None) -> None:
    """
    Write a Phantom dump file from a SarracenDataFrame.

    Both particle data and global header parameters stored in the sdf.params
    dict are written. If required Phantom parameters are missing, then Sarracen
    will automatically infer and set critical parameters such as the number of
    particles per particle type and the mass of each particle type.

    All entries in the sdf.params dict are written to the dump file. Phantom
    ignores parameters it does not recognise, so this is ok.

    By design, Sarracen does not invent values for parameters that are absent.
    Phantom typically warns about missing parameters and then chooses default
    values. For example, Sarracen does not require a *'time'* entry in
    sdf.params even though it sets the simulation time. If absent, then Phantom
    assumes t=0.

    Many parameters stored in the global header of the dump file are also
    specified in the Phantom .in file. The equation of state is one example.
    Phantom stores this in both the .in file and the header of the dump file
    via *'ieos'* (which can be set through the sdf.params dict).

    Parameters
    ----------
    filename : str
        Path to output file.
    sdf : SarracenDataFrame
        The particle data to write. Must contain positions (x, y, z) and
        smoothing lengths. Must have params dictionary set with required
        parameters.
    sinks : SarracenDataFrame, optional
        Sink particle data to write. Default is None.

    Raises
    ------
    ValueError
        If sdf contains NaN values.
        If sinks contains NaN values.
        If params is not set in sdf.
        If sdf is missing x, y, or z positions or smoothing lengths.
        If essential params are missing for the equation of state.

    See Also
    --------
    read_phantom : Read Phantom dump files.

    Notes
    -----
    If *'ieos'* is set, then Sarracen checks for additional parameters required
    for *'ieos'* = [1, 2, 3], corresponding to the isothermal, adiabatic, and
    locally isothermal equations of state.

    For periodic boundary conditions, set the domain limits using *{'xmin',
    'xmax', 'ymin', 'ymax', 'zmin', 'zmax'}*. If absent, then Phantom assumes a
    default periodic domain of x, y, z = [0, 1].

    To set physical units, set *{'udist', 'utime', 'umass', 'umagfd'}*.
    Otherwise, Phantom will assume default units.

    Examples
    --------
    Sarracen works well as a utility to modify existing Phantom dumps
    (moddump). This example shows how to use Sarracen to add a new sink
    particle to an existing Phantom dump.

    >>> import sarracen
    >>> sdf, sdf_sinks = sarracen.read_phantom('disc_00000')

    There is one existing sink particle at (0, 0, 0) with unit mass.

    >>> sdf_sinks
         x    y    z    m    h  ...   vx   vy   vz
    0  0.0  0.0  0.0  1.0  1.0  ...  0.0  0.0  0.0

    Create new sink properties and add to sinks SarracenDataFrame. Columns
    that were not specified will be NaN, which are set to 0 using fillna()
    in this example.

    >>> new = {'x': -1, 'y': 0, 'z': 0, 'm': 9.54e-4, 'h': 0.17, 'vy': -0.316}
    >>> sdf_sinks.loc[1] = new
    >>> sdf_sinks.fillna(0, inplace=True)
    >>> sdf_sinks
         x    y    z         m     h  ...   vx     vy   vz
    0  0.0  0.0  0.0  1.000000  1.00  ...  0.0  0.000  0.0
    1 -1.0  0.0  0.0  0.000954  0.17  ...  0.0 -0.316  0.0

    Then write the new Phantom dump file with the added sink particle.

    >>> sarracen.write_phantom('my_new_disc_00000', sdf, sdf_sinks)

    Note that this will copy all entries in the sdf.params dict to the new dump
    file. This is often useful when modifying existing dump files.

    """
    if sdf.isnull().values.any():
        raise ValueError("particle DataFrame contains NaNs or missing values.")

    if sinks is not None and sinks.isnull().values.any():
        raise ValueError("sinks DataFrame contains NaNs or missing values.")

    if sdf.params is None:
        raise ValueError("Parameters are not set in this SarracenDataFrame.")

    _check_for_essential_data(sdf)

    def_int = sdf.params.get('def_int_dtype', np.int32)
    def_real = sdf.params.get('def_real_dtype', np.float64)

    file = _write_capture_pattern(def_int, def_real)
    file += _write_file_identifier(sdf)
    file += _write_global_header(sdf, def_int, def_real)
    file += _write_array_blocks(sdf, def_int, def_real, sinks)

    with open(filename, 'wb') as phantom_file:
        phantom_file.write(file)
        phantom_file.close()



from typing import IO, List, Union
import pandas as pd
import numpy as np


def _determine_column_labels(file: IO) -> List[str]:
    """
    Determine the column labels.

    We need to read through the lines starting with '#' to find the specific
    line that contains the column labels. We assume it is the last line.
    """

    last_line = None

    while True:
        pos = file.tell()
        line = file.readline()
        if not line:
            raise ValueError("No data found in file.")

        # only lines starting '#' are valid
        if line.startswith("#"):
            last_line = line.strip()
        else:
            # skipped too far, rewind one line
            file.seek(pos)
            break

    if last_line is None:
        raise ValueError("No header line found in file.")

    # Get header labels
    # Assumes labels are exactly 3 spaces apart.
    # Assumes labels start with 2 digits (which we discard).
    # Seems like brittle code.
    # But this works. If this fails in future, can try switching to regex.
    line = last_line
    line = line[3:-1]
    labels = [x[2:].strip() for x in line.split(']   [')]

    return labels


def _infer_type(s: str) -> Union[np.float64, np.int32, str]:
    """
    Given a value from the ev, determine its type.

    Phantom writes floats using scientific notation always. Consider anything
    with a decimal point and an 'e' as a float. Otherwise it is an int.

    Special case is an integer with leading zeros. These are considered a str,
    as this is most likely the 'dump' column of an APR .ev file.
    """

    if '.' in s and 'e' in s.lower():
        try:
            return np.float64(s)
        except ValueError:
            pass

    if s.startswith('0') and s.isdigit():
        return s

    try:
        return np.int32(s)
    except ValueError:
        return s


def read_phantom_ev(filename: str) -> pd.DataFrame:
    """
    Read a Phantom .ev file and return a pandas DataFrame of the data.

    Each row of the DataFrame corresponds to a single line of the .ev file.
    The reader will skip any header lines that start with '#'. It assumes the
    last header line contains the column names. Values are generally stored
    as np.float64, but the reader will attempt to identify str and int values.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.

    Returns
    -------
    pandas DataFrame
    """

    data = []

    with open(filename) as file:

        labels = _determine_column_labels(file)

        # Read data
        for line in file:
            parts = line.strip().split()
            row = [_infer_type(x) for x in parts]
            data.append(row)

    return pd.DataFrame(data, columns=labels)



from ..kernels.base_kernel import BaseKernel
from ..kernels.cubic_spline import CubicSplineKernel
from ..kernels.quartic_spline import QuarticSplineKernel
from ..kernels.quintic_spline import QuinticSplineKernel

__all__ = ["BaseKernel", "CubicSplineKernel", "QuarticSplineKernel",
           "QuinticSplineKernel"]



from ..interpolate.base_backend import BaseBackend
from ..interpolate.cpu_backend import CPUBackend
from ..interpolate.gpu_backend import GPUBackend
from ..interpolate.interpolate import interpolate_2d_line, interpolate_2d, \
    interpolate_3d_line, interpolate_3d_proj, interpolate_3d_cross, \
    interpolate_3d_vec, interpolate_3d_cross_vec, interpolate_3d_grid, \
    interpolate_2d_vec

__all__ = ["BaseBackend", "CPUBackend", "GPUBackend",
           "interpolate_2d_line", "interpolate_2d", "interpolate_3d_line",
           "interpolate_3d_proj", "interpolate_3d_cross",
           "interpolate_3d_vec", "interpolate_3d_cross_vec",
           "interpolate_3d_grid", "interpolate_2d_vec"]



import numpy as np
import pandas as pd

from ..sarracen_dataframe import SarracenDataFrame


def classify_bound_particles(sdf: SarracenDataFrame,
                             sdf_sinks: SarracenDataFrame) -> pd.Series:
    """
    Determines to which sink each SPH particle is gravitationally bound.

    Returns a pandas Series that classifies particles by the sink to which they
    are bound, or -1 if they are not bound to any sink.

    Parameters
    ----------
    sdf : SarracenDataFrame
        The particle data.
    sdf_sinks : SarracenDataFrame
        The sink particles.

    Returns
    -------
    Series
        A pandas Series with the sink classification for each particle.

    Raises
    ------
    ValueError
        If the particle SarracenDataFrame does not contain particle mass.

    Notes
    -----
    The index of the Series will match the index of the input
    SarracenDataFrame. The sinks are classified according to their index in the
    sdf_sinks.

    Examples
    --------
    Here is a set of SPH particles and sink particles. There are 4 sink
    particles.

    >>> sdf, sdf_sinks = sarracen.read_phantom('dumpfile')
    >>> len(sdf_sinks)
    4
    >>> sdf.head()
              x         y        z      vx      vy      vz       h
    0   24.0711   34.7009  -9.7994 -0.1229  0.0843  0.0012  1.2793
    1  -16.8095   39.8220  10.3250 -0.1356 -0.0585 -0.0010  1.3448
    2  -31.4485  143.1688  33.4715 -0.0747 -0.0158  0.0002  3.1289
    3   -2.0287 -149.1813 -35.7251  0.0749 -0.0012 -0.0011  3.2898
    4  -93.4897   85.5293   2.2384 -0.0559 -0.0629  0.0006  1.9655

    A pandas Series is returned that classifies which sink particle each SPH
    particle is bound to.

    >>> bound_to_sink = sarracen.ptmass.classify_sink(sdf, sdf_sinks)
    >>> sdf['sink'] = bound_to_sink
    >>> sdf.head()
               x         y       z      vx      vy      vz       h  sink
    0   24.0711   34.7009  -9.7994 -0.1229  0.0843  0.0012  1.2793     0
    1  -16.8095   39.8220  10.3250 -0.1356 -0.0585 -0.0010  1.3448     0
    2  -31.4485  143.1688  33.4715 -0.0747 -0.0158  0.0002  3.1289     3
    3   -2.0287 -149.1813 -35.7251  0.0749 -0.0012 -0.0011  3.2898     2
    4  -93.4897   85.5293   2.2384 -0.0559 -0.0629  0.0006  1.9655     0

    """
    # calculate the energy of particles relative to each sink

    if sdf_sinks.empty:
        raise ValueError("Sink particle DataFrame is empty.")

    if {sdf.mcol}.issubset(sdf.columns):
        mass = sdf[sdf.mcol]
    else:
        if sdf.params is None or 'mass' not in sdf.params:
            raise ValueError("Cannot find particle mass in "
                             "the SarracenDataFrame.")
        mass = sdf.params['mass']

    x = sdf[sdf.xcol].to_numpy(copy=False)
    y = sdf[sdf.ycol].to_numpy(copy=False)
    z = sdf[sdf.zcol].to_numpy(copy=False)
    vx = sdf[sdf.vxcol].to_numpy(copy=False)
    vy = sdf[sdf.vycol].to_numpy(copy=False)
    vz = sdf[sdf.vzcol].to_numpy(copy=False)

    energy = pd.Series(np.inf, index=sdf.index)
    bound_sink = pd.Series(-1, index=sdf.index, dtype=sdf_sinks.index.dtype)

    for sink_idx, sink_row in sdf_sinks.iterrows():

        x_sink = sink_row[sdf_sinks.xcol]
        y_sink = sink_row[sdf_sinks.ycol]
        z_sink = sink_row[sdf_sinks.zcol]
        vx_sink = sink_row[sdf_sinks.vxcol]
        vy_sink = sink_row[sdf_sinks.vycol]
        vz_sink = sink_row[sdf_sinks.vzcol]

        v_relsq = (np.square(vx - vx_sink)
                   + np.square(vy - vy_sink)
                   + np.square(vz - vz_sink))
        r = np.sqrt(np.square(x - x_sink)
                    + np.square(y - y_sink)
                    + np.square(z - z_sink))

        # sum of kinetic and potential energies
        E_total = 0.5 * mass * v_relsq - sink_row[sdf_sinks.mcol] * mass / r

        update = E_total < energy
        energy[update] = E_total[update]
        bound_sink[update] = sink_idx

    bound_sink[energy > 0.0] = -1

    return bound_sink
